import scipy.sparse
import scipy.sparse.linalg
from scipy.special import sph_harm
from scipy.special import binom
from scipy.special import factorial
import numpy as np

"""
Wigner 3j-symbol and Clebsch-Gordon coefficients
"""

def fac(n):
    if n < 0:
        return 0
    return float(np.math.factorial(n))

def Yslm(s, l, m, th):
    if np.abs(s) > l:
        return 0.*th
    if s == 0:
        return np.real(sph_harm(m, l, 0., th))
    elif s + m < 0:
        return (-1.)**(s+m)*YslmBase(-s, l, -m, np.cos(th))
#     elif th > np.pi/2.:
#         return (-1.)**(l + m)*YslmBase(-s, l, m, -np.cos(th))
    else:
        return YslmBase(s, l, m, np.cos(th))

def YslmBase(s, l, m, z):
    rmax = l - s
    pref = (0.5)**(l)*(-1.)**m*np.sqrt(factorial(l+m)/factorial(l+s)*factorial(l-m)/factorial(l-s)*(2*l+1)/(4.*np.pi))*np.sqrt(1. - z)**(s + m)*np.sqrt(1. + z)**(s - m)
    
    yslm = 0.*pref
    for r in range(0, rmax + 1):
        yslm += binom(l - s, r)*binom(l + s, r + s - m)*(z - 1.)**(rmax - r)*(z + 1.)**(r)
    
    return pref*yslm

def clebsch(l1, l2, l3, m1, m2, m3):
    return (-1)**(l1 - l2 + m3)*np.sqrt(2*l3 + 1)*w3j(l1, l2, l3, m1, m2, -m3);

def w3j(l1, l2, l3, m1, m2, m3):
    if m1 + m2 + m3 != 0:
        return 0
    elif abs(l1 - l2) > l3:
        return 0
    elif l1 + l2 < l3:
        return 0
    
    if abs(m1) > l1:
        return 0
    elif abs(m2) > l2:
        return 0
    elif abs(m3) > l3:
        return 0
    
    sumTerm = w3j_tsum(l1, l2, l3, m1, m2, m3)
    if sumTerm == 0:
        return 0
    sumSign = np.sign(sumTerm)
    tempLog = 0.5*(np.log(fac(l1 + m1)) + np.log(fac(l2 + m2)) + np.log(fac(l3 + m3)))
    tempLog += 0.5*(np.log(fac(l1 - m1)) + np.log(fac(l2 - m2)) + np.log(fac(l3 - m3)))
    tempLog += np.log(triangle_coeff(l1, l2, l3))
    tempLog += np.log(abs(sumTerm))
    
    temp = sumSign*np.exp(tempLog)
    temp *= (-1)**(l1-l2-m3)
    
    return temp
    
def w3j_tsum(l1, l2, l3, m1, m2, m3):
    t_min_num = w3j_t_min(l1, l2, l3, m1, m2, m3)
    t_max_num = w3j_t_max(l1, l2, l3, m1, m2, m3)
    x = 0
    if t_max_num < t_min_num:
        t_max_num = t_min_num

    for t in range(t_min_num - 1, t_max_num + 2):
        term = (fac(t)*fac(l3 - l2 + m1 + t)*fac(l3 - l1 - m2 + t)
            *fac(l1 + l2 - l3 - t)*fac(l1 - t - m1)*fac(l2 - t + m2))
        if term > 0:
            x += (-1)**t/term
    
    return x

def w3j_t_min(l1, l2, l3, m1, m2, m3):
    temp = 0
    
    comp = l3 - l2 + m1
    if temp + comp < 0:
        temp = -comp
    comp = l3 - l1 - m2
    if temp + comp < 0:
        temp = -comp
        
    return temp

def w3j_t_max(l1, l2, l3, m1, m2, m3):
    temp = 1
    comp = l1 + l2 - l3
    if comp - temp > 0:
        temp = comp
    comp = l1 - m1
    if comp - temp > 0:
        temp = comp
    comp = l2 + m2
    if comp - temp > 0:
        temp = comp
        
    return temp;

def triangle_coeff(l1, l2, l3):
    return np.sqrt(fac(l1 + l2 - l3)*fac(l3 + l1 - l2)*fac(l2 + l3 - l1)/fac(l1 + l2 + l3 + 1))

"""
SWSH Eigenvalue Functions
"""

def k1(s, l, j, m):
    return np.sqrt((2*l + 1)/(2*j + 1))*clebsch(l, 1, j, m, 0, m)*clebsch(l, 1, j, -s, 0, -s);

def k2(s, l, j, m):
    ktemp = 2./3.*np.sqrt((2*l + 1)/(2*j + 1))*clebsch(l, 2, j, m, 0, m)*clebsch(l, 2, j, -s, 0, -s);
    if j == l:
        ktemp += 1/3.
    return ktemp

def k2m2(s, l, m):
    temp = (l - m - 1.)/(l - 1.)
    temp *= (l + m - 1.)/(l - 1.)
    temp *= np.float64(l - m)/l
    temp *= np.float64(l + m)/l
    temp *= (l - s)/(2.*l - 1.)
    temp *= (l + s)/(2.*l - 1.)
    temp *= (l - s - 1.)/(2.*l + 1.)
    temp *= (l + s - 1.)/(2.*l - 3.)
    return np.sqrt(temp)

def k2m1(s, l, m):
    temp = np.float64(l - m)*np.float64(l + m)/(2.*l - 1.)
    temp *= np.float64(l - s)*np.float64(l + s)/(2.*l + 1.)
    return -2.*m*s*np.sqrt(temp)/l/(l - 1.)/(l + 1.)

def k2p0(s, l, m):
    temp = np.float64(l*(l + 1.) - 3.*m*m)/(2.*l - 1.)/l
    temp *= np.float64(l*(l + 1.) - 3.*s*s)/(2.*l + 3.)/(l + 1.)
    return 1./3.*(1. + 2.*temp)

def k2p1(s, l, m):
    temp = np.float64(l - m + 1.)*np.float64(l + m + 1.)/(2.*l + 1.)
    temp *= np.float64(l - s + 1.)*np.float64(l + s + 1.)/(2.*l + 3.)
    return -2.*m*s*np.sqrt(temp)/l/(l + 1.)/(l + 2.)

def k2p2(s, l, m):
    temp = (l - m + 1.)/(l + 1.)
    temp *= (l + m + 1.)/(l + 1.)
    temp *= (l - m + 2.)/(l + 2.)
    temp *= (l + m + 2.)/(l + 2.)
    temp *= (l - s + 2.)/(2.*l + 3.)
    temp *= (l + s + 2.)/(2.*l + 3.)
    temp *= (l - s + 1.)/(2.*l + 1.)
    temp *= (l + s + 1.)/(2.*l + 5.)
    return np.sqrt(temp)

def k1m1(s, l, m):
    temp = np.float64(l - m)*np.float64(l + m)/(2.*l - 1.)
    temp *= np.float64(l - s)*np.float64(l + s)/(2.*l + 1.)
    return np.sqrt(temp)/l

def k1p0(s, l, m):
    return -np.float64(m*s)/l/(l + 1.)

def k1p1(s, l, m):
    temp = np.float64(l - m + 1.)*np.float64(l + m + 1.)/(2.*l + 3.)
    temp *= np.float64(l - s + 1.)*np.float64(l + s + 1.)/(2.*l + 1.)
    return np.sqrt(temp)/(l + 1.)

def akm2(s, l, m, g):
    return -g*g*k2m2(s, l, m)

def akm1(s, l, m, g):
    return -g*g*k2m1(s, l, m) + 2.*s*g*k1m1(s, l, m)

def akp0(s, l, m, g):
    return -g*g*k2p0(s, l, m) + 2.*s*g*k1p0(s, l, m) + l*(l + 1.) - s*(s + 1.) - 2.*m*g + g*g

def akp1(s, l, m, g):
    return -g*g*k2p1(s, l, m) + 2.*s*g*k1p1(s, l, m)

def akp2(s, l, m, g):
    return -g*g*k2p2(s, l, m)

def spectral_sparse_matrix(s, m, g, nmax):
    lmin = max(abs(s), abs(m))
    larray = np.arange(lmin, lmin + nmax)
    return scipy.sparse.diags([akm2(s, larray[2:], m, g), akm1(s, larray[1:], m, g), akp0(s, larray, m, g), akp1(s, larray[:-1], m, g), akp2(s, larray[:-2], m, g)], [-2, -1, 0, 1, 2])

def swsh_eigs(s, l, m, g, nmax=None, return_eigenvectors=True):
    lmin = max(abs(s), abs(m))
    kval = l - lmin
    
    if nmax is None:
        buffer = round(20 + 2*g)
        Nmax = kval + buffer + 2
    else:
        if nmax < kval:
            Nmax = kval + 5
        else:
            Nmax = nmax
    
    mat = spectral_sparse_matrix(s, m, g, Nmax)
    out = scipy.sparse.linalg.eigs(mat, k=Nmax-2, which='SM', return_eigenvectors=return_eigenvectors)
    
    return out

def Yslm_eigenvalue(s, l, *args):
    return l*(l + 1.) - s*(s + 1.)

def swsh_eigenvalue(s, l, m, g, nmax=None):
    if g == 0.:
        return Yslm_eigenvalue(s, l)
    
    las = swsh_eigs(s, l, m, g, nmax=nmax, return_eigenvectors=False)
    
    return np.real(las[::-1][l - max(abs(s), abs(m))])

def swsh_coeffs(s, l, m, g, th):
    if g == 0.:
        return Yslm(s, l, m, th)
    
    _, eig = swsh_eigs(s, l, m, g, nmax=None, return_eigenvectors=True)
    coeffs = np.real(eig[l - max(abs(s), abs(m))])
    
    return np.real(eig[l - max(abs(s), abs(m))])
        
class SWSHBase:
    def __init__(self, *args):
        arg_num = np.array(args).shape[0]
        if arg_num < 3:
            print('Error. Not enough arguments to create class')
            pass
        
        self.s = args[0]
        self.l = args[1]
        self.m = args[2]
        self.lmin = max(abs(self.s), abs(self.m))

        if arg_num > 3:
            self.spheroidicity = args[3]
            
class SWSHSeriesBase(SWSHBase):
    def __init__(self, s, l, m, g):
        SWSHBase.__init__(self, s, l, m, g)
        
    def sparse_matrix(self, nmax):
        return spectral_sparse_matrix(self.s, self.m, self.spheroidicity, nmax)
        
    def eigs(self, nmax = None, **kwargs):
        kval = self.l - self.lmin
    
        if nmax is None:
            buffer = round(20 + 2*self.spheroidicity)
            Nmax = kval + buffer + 2
        else:
            if nmax < kval:
                Nmax = kval + 5
            else:
                Nmax = nmax
                
        if "k" not in kwargs.keys():
            kwargs["k"] = Nmax - 2
                
        if "which" not in kwargs.keys():
            kwargs["which"] = 'SM'
            
#         if "sigma" not in kwargs.keys():
#             kwargs["sigma"] = (self.l + Nmax)*(self.l + Nmax + 1) - self.s*(self.s + 1)
            
        if "return_eigenvectors" not in kwargs.keys():
            kwargs["return_eigenvectors"] = True
            
        mat = self.sparse_matrix(Nmax)
        return scipy.sparse.linalg.eigs(mat, **kwargs)
    
    def generate_eigenvalue(self):
        las = np.real(self.eigs(return_eigenvectors=False))
        pos = np.argsort(las)[self.l - self.lmin]
        return las[pos]
    
    def generate_eigs(self):
        las, eigs = self.eigs()
        pos = np.argsort(np.real(las))[self.l - self.lmin]
        eigs_temp = np.real(eigs[:, pos])
        eigs_return = np.sign(eigs_temp[self.l - self.lmin])*eigs_temp
        return (np.real(las[pos]), eigs_return)
    
class SWSH(SWSHSeriesBase):
    def __init__(self, s, l, m, g):
        SWSHSeriesBase.__init__(self, s, l, m, g)
        if self.spheroidicity == 0.:
            self.eval = self.Yslm
            self.eigenvalue = Yslm_eigenvalue(self.s, self.l)
            self.coeffs = np.zeros(self.l - self.lmin)
            self.coeffs[-1] = 1.
        else:
            self.eval = self.Sslm
            self.eigenvalue, self.coeffs = self.generate_eigs()
            
    def Yslm(self, l, th):
        return Yslm(self.s, l, self.m, th)
    
    def Sslm(self, *args):
        th = args[-1]
        term_num = self.coeffs.shape[0]
        pts_num = th.shape[0]
        Yslm_array = np.empty((term_num, pts_num))
        for i in range(term_num):
            Yslm_array[i] = self.Yslm(self.lmin + i, th)
            
        return np.dot(self.coeffs, Yslm_array)
            
    def __call__(self, th):
        return self.eval(self.l, th)
    
def muCoupling(s, l):
    """
    Eigenvalue for the spin-weighted spherical harmonic lowering operator
    Setting s -> -s gives the negative of the eigenvalue for the raising operator
    """
    if l + s < 0 or l - s + 1. < 0:
        return 0
    return np.sqrt((l - s + 1.)*(l + s))

def Asjlm(s, j, l, m):
    if s >= 0:
        return (-1.)**(m + s)*np.sqrt(4**s*fac(s)**2*(2*l + 1)*(2*j + 1)/fac(2*s))*w3j(s, l, j, 0, m, -m)*w3j(s, l, j, s, -s, 0)
    else:
        return (-1.)**(m)*np.sqrt(4**(-s)*fac(-s)**2*(2*l + 1)*(2*j + 1)/fac(-2*s))*w3j(-s, l, j, 0, m, -m)*w3j(-s, l, j, s, -s, 0)

def spin_operator_normalization(s, ns, l):
    s_sgn = np.sign(s)
    nmax1 = np.abs(s) + 1
    Jterm = 1.
    for ni in range(1, ns + 1):
        Jterm *= -s_sgn*muCoupling((nmax1-ni), l)
    return Jterm
        


# from geo_wrap cimport GeodesicSource, kerr_geo_orbit, kerr_geo_orbital_constants, kerr_geo_mino_frequencies

from libcpp.vector cimport vector
import numpy as np
cimport numpy as np

from libcpp.vector cimport vector

cdef extern from "geo.hpp":
    cdef cppclass GeodesicTrajectory:
        GeodesicTrajectory()
        GeodesicTrajectory(vector[double] tR, vector[double] tTheta, vector[double] r, vector[double] theta, vector[double] phiR, vector[double] phiTheta)
        vector[double] tR, tTheta, r, theta, phiR, phiTheta

    cdef cppclass GeodesicConstants:
        GeodesicConstants()
        GeodesicConstants(double a, double p, double e, double x, double En, double Lz, double Q)
        GeodesicConstants(double a, double p, double e, double x, double En, double Lz, double Q, double, double, double, double)
        GeodesicConstants(double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double)
        double a, p, e, x
        double En, Lz, Q
        double r1, r2, r3, r4, z1, z2
        double upsilonT, upsilonR, upsilonTheta, upsilonPhi
        double carterR, carterTheta, carterPhi

    cpdef cppclass GeodesicSource:
        # GeodesicSource() except +
        GeodesicSource(double a, double p, double e, double x, int Nsample) except +
        # ~GeodesicSource()

        int getOrbitalSampleNumber()
        double getBlackHoleSpin()
        double getSemiLatusRectum()
        double getEccentricity()
        double getInclination()
        double getOrbitalEnergy()
        double getOrbitalAngularMomentum()
        double getCarterConstant()
        double getRadialRoot(int i)
        double getPolarRoot(int i)
        double getMinoFrequency(int mu)
        double getTimeFrequency(int i)
        double getTimeFrequency(int m, int k, int n)
        double getCarterFrequency(int i)
        double getCarterFrequency(int m, int k, int n)

        vector[double] getTimeAccumulation(int j)
        vector[double] getRadialPosition()
        vector[double] getPolarPosition()
        vector[double] getAzimuthalAccumulation(int j)

        double getTimeAccumulation(int j, int pos)
        double getRadialPosition(int pos)
        double getPolarPosition(int pos)
        double getAzimuthalAccumulation(int j, int pos)

        double getTimePositionOfMinoTime(double la)
        double getRadialPositionOfMinoTime(double la)
        double getPolarPositionOfMinoTime(double la)
        double getAzimuthalPositionOfMinoTime(double la)

        vector[double] getTimePositionOfMinoTime(vector[double] la)
        vector[double] getRadialPositionOfMinoTime(vector[double] la)
        vector[double] getPolarPositionOfMinoTime(vector[double] la)
        vector[double] getAzimuthalPositionOfMinoTime(vector[double] la)
        vector[double] getPositionOfMinoTime(vector[double] la)

        double getMinoTimeOfTime(double t)

        vector[double] getTimeCoefficients(int j)
        vector[double] getRadialCoefficients()
        vector[double] getPolarCoefficients()
        vector[double] getAzimuthalCoefficients(int j)

        GeodesicConstants getConstants()
        GeodesicTrajectory getTrajectory()
        GeodesicTrajectory getCoefficients()

    GeodesicSource kerr_geo_orbit(double a, double p, double e, double x, int n)
    void kerr_geo_orbital_constants(double &En, double &Lz, double &Qc, double &a, double &p, double &e, double &x)
    void kerr_geo_radial_roots(double &r1, double &r2, double &r3, double &r4, double &a, double &p, double &e, double &En, double &Lz, double &Qc)
    void kerr_geo_polar_roots(double &z1, double &z2, double &a, double &x, double &En, double &Lz, double &Qc)
    void kerr_geo_mino_frequencies(double &upT, double &upR, double &upTh, double &upPh, double &a, double &p, double &e, double &x)

    double kerr_geo_VtR(const double & a, const double & En, const double & Lz, const double & Q, const double & r)
    double kerr_geo_VtTheta(const double & a, const double & En, const double & Lz, const double & Q, const double & theta)
    double kerr_geo_Vr(const double & a, const double & En, const double & Lz, const double & Q, const double & r)
    double kerr_geo_Vtheta(const double & a, const double & En, const double & Lz, const double & Q, const double & theta)
    double kerr_geo_VphiR(const double & a, const double & En, const double & Lz, const double & Q, const double & r)
    double kerr_geo_VphiTheta(const double & a, const double & En, const double & Lz, const double & Q, const double & theta)
    double kerr_geo_Vz(const double & a, const double & En, const double & Lz, const double & Q, const double & z)
    double kerr_geo_Vz_dz(const double & a, const double & En, const double & Lz, const double & Q, const double & z)
    double kerr_geo_Vz_dz2(const double & a, const double & En, const double & Lz, const double & Q, const double & z)
    double kerr_isco(double a, int sgnX)
    double kerr_isco_frequency(double a)

def kerr_geo_V01(double a, double En, double Lz, double Q, double r):
    return kerr_geo_VtR(a, En, Lz, Q, r)

def kerr_geo_V02(double a, double En, double Lz, double Q, double theta):
    return kerr_geo_VtTheta(a, En, Lz, Q, theta)

def kerr_geo_V11(double a, double En, double Lz, double Q, double r):
    return kerr_geo_Vr(a, En, Lz, Q, r)

def kerr_geo_V22(double a, double En, double Lz, double Q, double theta):
    return kerr_geo_Vtheta(a, En, Lz, Q, theta)

def kerr_geo_V31(double a, double En, double Lz, double Q, double r):
    return kerr_geo_VphiR(a, En, Lz, Q, r)

def kerr_geo_V32(double a, double En, double Lz, double Q, double theta):
    return kerr_geo_VphiTheta(a, En, Lz, Q, theta)

cdef class KerrGeodesic:
    cdef GeodesicSource *geocpp
    cdef int nsamplescpp

    def __init__(self, double a, double p, double e, double x, int nsamples = 2**8):
        self.geocpp = new GeodesicSource(a, p, e, x, nsamples)
        self.nsamplescpp = nsamples

    def __dealloc__(self):
        del self.geocpp

    @property
    def nsamples(self):
        return self.nsamplescpp

    @property
    def blackholespin(self):
        return self.geocpp.getBlackHoleSpin()
    
    @property
    def semilatusrectum(self):
        return self.geocpp.getSemiLatusRectum()
    
    @property
    def eccentricity(self):
        return self.geocpp.getEccentricity()

    @property
    def inclination(self):
        return self.geocpp.getInclination()

    @property
    def orbitalenergy(self):
        return self.geocpp.getOrbitalEnergy()

    @property
    def orbitalangularmomentum(self):
        return self.geocpp.getOrbitalAngularMomentum()

    @property
    def carterconstant(self):
        return self.geocpp.getCarterConstant()

    @property
    def radialroots(self):
        return np.array([self.geocpp.getRadialRoot(1+i) for i in range(4)])

    @property
    def polarroots(self):
        return np.array([self.geocpp.getPolarRoot(1+i) for i in range(2)])

    @property
    def minofrequencies(self):
        return np.array([self.geocpp.getMinoFrequency(i) for i in range(4)])

    @property
    def timefrequencies(self):
        return np.array([self.geocpp.getTimeFrequency(i) for i in range(1, 4)])

    @property
    def frequencies(self):
        return self.timefrequencies

    @property
    def carterfrequencies(self):
        return np.array([self.geocpp.getCarterFrequency(i) for i in range(1, 4)]) 

    def mode_time_frequency(self, np.ndarray[ndim=1, dtype=np.int64_t] kvec):
        return np.dot(kvec, (self.frequencies))
    
    mode_frequency = mode_time_frequency

    def mode_carter_frequency(self, np.ndarray[ndim=1, dtype=np.int64_t] kvec):
        return np.dot(kvec,(self.carterfrequencies))

    cdef void getTimePositionOfMinoTimeArray(self, np.float64_t *t, np.float64_t *la, int n):
        for i in range(n):
            t[i] = self.geocpp.getTimePositionOfMinoTime(la[i])
    cdef void getRadialPositionOfMinoTimeArray(self, np.float64_t *t, np.float64_t *la, int n):
        for i in range(n):
            t[i] = self.geocpp.getRadialPositionOfMinoTime(la[i])
    cdef void getPolarPositionOfMinoTimeArray(self, np.float64_t *t, np.float64_t *la, int n):
        for i in range(n):
            t[i] = self.geocpp.getPolarPositionOfMinoTime(la[i])
    cdef void getAzimuthalPositionOfMinoTimeArray(self, np.float64_t *t, np.float64_t *la, int n):
        for i in range(n):
            t[i] = self.geocpp.getAzimuthalPositionOfMinoTime(la[i])
    
    def time_position(self, np.ndarray[ndim=1, dtype=np.float64_t] la):
        cdef int n = la.shape[0]
        cdef np.ndarray[ndim=1, dtype=np.float64_t] t = np.empty(n, dtype = np.float64)
        self.getTimePositionOfMinoTimeArray(&t[0], &la[0], n)
        # for some reason this is marginally slower than the line above with the cdef function
        # for i in range(n):
        #     t[i] = self.geocpp.getTimePositionOfMinoTime(la[i])
        return t
    
    def radial_position(self, np.ndarray[ndim=1, dtype=np.float64_t] la):
        cdef int n = la.shape[0]
        cdef np.ndarray[ndim=1, dtype=np.float64_t] x = np.empty(n, dtype = np.float64)
        self.getRadialPositionOfMinoTimeArray(&x[0], &la[0], n)
        return x

    def polar_position(self, np.ndarray[ndim=1, dtype=np.float64_t] la):
        cdef int n = la.shape[0]
        cdef np.ndarray[ndim=1, dtype=np.float64_t] x = np.empty(n, dtype = np.float64)
        self.getPolarPositionOfMinoTimeArray(&x[0], &la[0], n)
        return x

    def azimuthal_position(self, np.ndarray[ndim=1, dtype=np.float64_t] la):
        cdef int n = la.shape[0]
        cdef np.ndarray[ndim=1, dtype=np.float64_t] x = np.empty(n, dtype = np.float64)
        self.getAzimuthalPositionOfMinoTimeArray(&x[0], &la[0], n)
        return x
    
    def get_time_accumulation(self, int j):
        cdef vector[double] deltaX_cpp = self.geocpp.getTimeAccumulation(j)
        cdef int n = deltaX_cpp.size()
        cdef np.ndarray[ndim=1, dtype=np.float64_t] deltaX = np.empty(n, dtype = np.float64)
        for i in range(n):
            deltaX[i] = deltaX_cpp[i]
        return deltaX

    def get_radial_points(self):
        cdef vector[double] deltaX_cpp = self.geocpp.getRadialPosition()
        cdef int n = deltaX_cpp.size()
        cdef np.ndarray[ndim=1, dtype=np.float64_t] deltaX = np.empty(n, dtype = np.float64)
        for i in range(n):
            deltaX[i] = deltaX_cpp[i]
        return deltaX

    def get_polar_points(self):
        cdef vector[double] deltaX_cpp = self.geocpp.getPolarPosition()
        cdef int n = deltaX_cpp.size()
        cdef np.ndarray[ndim=1, dtype=np.float64_t] deltaX = np.empty(n, dtype = np.float64)
        for i in range(n):
            deltaX[i] = deltaX_cpp[i]
        return deltaX
    
    def get_azimuthal_accumulation(self, int j):
        cdef vector[double] deltaX_cpp = self.geocpp.getAzimuthalAccumulation(j)
        cdef int n = deltaX_cpp.size()
        cdef np.ndarray[ndim=1, dtype=np.float64_t] deltaX = np.empty(n, dtype = np.float64)
        for i in range(n):
            deltaX[i] = deltaX_cpp[i]
        return deltaX

    def position(self, double la):
        return np.array([self.geocpp.getTimePositionOfMinoTime(la), self.geocpp.getRadialPositionOfMinoTime(la), self.geocpp.getPolarPositionOfMinoTime(la), self.geocpp.getAzimuthalPositionOfMinoTime(la)])
    
    def position_vec(self, np.ndarray[ndim=1, dtype=np.float64_t] la):
        cdef np.ndarray[ndim=2, dtype=np.float64_t] xp = np.empty((la.shape[0], 4), dtype=np.float64)
        for i in range(la.shape[0]):
            xp[i] = self.position(la[i])
        return xp.T

    def mino_time(self, double t):
        return self.geocpp.getMinoTimeOfTime(t)

    def get_time_coefficients(self, int j):
        cdef vector[double] deltaX_cpp = self.geocpp.getTimeCoefficients(j)
        cdef int n = deltaX_cpp.size()
        cdef np.ndarray[ndim=1, dtype=np.float64_t] deltaX = np.empty(n, dtype = np.float64)
        for i in range(n):
            deltaX[i] = deltaX_cpp[i]
        return deltaX

    def get_radial_coefficients(self):
        cdef vector[double] deltaX_cpp = self.geocpp.getRadialCoefficients()
        cdef int n = deltaX_cpp.size()
        cdef np.ndarray[ndim=1, dtype=np.float64_t] deltaX = np.empty(n, dtype = np.float64)
        for i in range(n):
            deltaX[i] = deltaX_cpp[i]
        return deltaX

    def get_polar_coefficients(self):
        cdef vector[double] deltaX_cpp = self.geocpp.getPolarCoefficients()
        cdef int n = deltaX_cpp.size()
        cdef np.ndarray[ndim=1, dtype=np.float64_t] deltaX = np.empty(n, dtype = np.float64)
        for i in range(n):
            deltaX[i] = deltaX_cpp[i]
        return deltaX
    
    def get_azimuthal_coefficients(self, int j):
        cdef vector[double] deltaX_cpp = self.geocpp.getAzimuthalCoefficients(j)
        cdef int n = deltaX_cpp.size()
        cdef np.ndarray[ndim=1, dtype=np.float64_t] deltaX = np.empty(n, dtype = np.float64)
        for i in range(n):
            deltaX[i] = deltaX_cpp[i]
        return deltaX


def kerr_orbital_constants_wrapper(double a, double p, double e, double x):
    cdef double En, Lz, Qc
    En = 0.
    Lz = 0.
    Qc = 0.
    kerr_geo_orbital_constants(En, Lz, Qc, a, p, e, x)
    return np.array([En, Lz, Qc])

def kerr_mino_frequencies_wrapper(double a, double p, double e, double x):
    cdef double upT, upR, upTh, upPhi
    upT = 0.
    upR = 0.
    upTh = 0.
    upPhi = 0.
    kerr_geo_mino_frequencies(upT, upR, upTh, upPhi, a, p, e, x)
    return np.array([upT, upR, upTh, upPhi])
        


from libcpp.vector cimport vector
from libcpp.complex cimport complex as cpp_complex
from cython.operator cimport dereference
import numpy as np
cimport numpy as np

# from geo_wrap cimport GeodesicSource
include "geo_wrap.pyx"
include "radialsolver_wrap.pyx"

cdef extern from "teukolsky.hpp":
    cdef cppclass SpinWeightedHarmonic:
        pass

    cdef cppclass TeukolskyModeCPP "TeukolskyMode":
        # TeukolskyModeCPP(int L, int m, int k, int n, GeodesicSource& geo)
        TeukolskyModeCPP(int s, int L, int m, int k, int n, GeodesicSource& geo)
        TeukolskyModeCPP(int s, int L, int m, int k, int n, double a, vector[double] theta, vector[double] r)

        int generateSolutions(GeodesicSource& geo, SolutionMethod method, int samplesize)
        int generateSolutions(double omega, GeodesicTrajectory &traj, GeodesicConstants &geoConst, vector[double] r, vector[double] theta, SolutionMethod method, int samplesize)
        int generateSolutions(SpinWeightedHarmonic swsh, RadialTeukolsky teuk, GeodesicTrajectory& traj, GeodesicConstants &geoConst)

        int flipSpinWeightAndFrequency()
        int flipSpinWeight()

        int getSpinWeight()
        int getSpheroidalModeNumber()
        int getAzimuthalModeNumber()
        int getPolarModeNumber()
        int getRadialModeNumber()
        int getSampleSize()
        double getBlackHoleSpin()
        double getFrequency()
        double getHorizonFrequency()
        double getEigenvalue()
        vector[double] getCouplingCoefficient()
        double getCouplingCoefficient(int l)
        int getMinCouplingModeNumber()
        int getMaxCouplingModeNumber()

        vector[double] getRadialPoints()
        cpp_complex[double] getHomogeneousRadialSolution(BoundaryCondition bc)
        cpp_complex[double] getHomogeneousRadialDerivative(BoundaryCondition bc)
        cpp_complex[double] getHomogeneousSecondRadialDerivative(BoundaryCondition bc)
        cpp_complex[double] getTeukolskyAmplitude(BoundaryCondition bc)
        double getTeukolskyAmplitudePrecision(BoundaryCondition bc)
        cpp_complex[double] getRadialSolution(BoundaryCondition bc)
        cpp_complex[double] getRadialDerivative(BoundaryCondition bc)

        vector[double] getPolarPoints()
        vector[double] getPolarSolution()
        vector[double] getPolarDerivative()

        int getRadialSampleNumber()
        double getRadialPoints(int pos)
        cpp_complex[double] getHomogeneousRadialSolution(BoundaryCondition bc, int pos)
        cpp_complex[double] getHomogeneousRadialDerivative(BoundaryCondition bc, int pos)
        cpp_complex[double] getHomogeneousSecondRadialDerivative(BoundaryCondition bc, int pos)
        cpp_complex[double] getRadialSolution(BoundaryCondition bc, int pos)
        cpp_complex[double] getRadialDerivative(BoundaryCondition bc, int pos)

        int getPolarSampleNumber()
        double getPolarPoints(int pos)
        double getPolarSolution(int pos)
        double getPolarDerivative(int pos)
        double getPolarSecondDerivative(int pos)

cdef extern from "hertz.hpp":
    cdef enum Gauge:
        ORG, IRG, SAAB0, SAAB4, ASAAB0, ASAAB4, SAAB, ASAAB

    cdef cppclass HertzModeCPP "HertzMode":
        HertzModeCPP(TeukolskyModeCPP& teuk, Gauge gauge)

        int generateSolutions()

        Gauge getGauge()
        int getSpinWeight()
        int getSpheroidalModeNumber()
        int getAzimuthalModeNumber()
        int getPolarModeNumber()
        int getRadialModeNumber()
        int getSampleSize()
        double getBlackHoleSpin()
        double getFrequency()
        double getHorizonFrequency()
        double getEigenvalue()

        # Vector getCouplingCoefficient()
        double getCouplingCoefficient(int l)
        int getMinCouplingModeNumber()
        int getMaxCouplingModeNumber()

        # Vector getScalarCouplingCoefficient()
        double getScalarCouplingCoefficient(int l)
        int getMinScalarCouplingModeNumber()
        int getMaxScalarCouplingModeNumber()

        # Vector getRadialPoints()
        # ComplexVector getHomogeneousRadialSolution(BoundaryCondition bc)
        # ComplexVector getHomogeneousRadialDerivative(BoundaryCondition bc)
        cpp_complex[double] getHertzAmplitude(BoundaryCondition bc)
        # ComplexVector getRadialSolution(BoundaryCondition bc)
        # ComplexVector getRadialDerivative(BoundaryCondition bc)

        # Vector getPolarPoints()
        # Vector getPolarSolution()
        # Vector getPolarDerivative()

        double getRadialPoints(int pos)
        cpp_complex[double] getHomogeneousRadialSolution(BoundaryCondition bc, int pos)
        cpp_complex[double] getHomogeneousRadialDerivative(BoundaryCondition bc, int pos)
        cpp_complex[double] getHomogeneousRadialSecondDerivative(BoundaryCondition bc, int pos)
        cpp_complex[double] getHomogeneousRadialSolution(BoundaryCondition bc, int dr, int pos)
        cpp_complex[double] getRadialSolution(BoundaryCondition bc, int pos)
        cpp_complex[double] getRadialDerivative(BoundaryCondition bc, int pos)
        cpp_complex[double] getRadialSecondDerivative(BoundaryCondition bc, int pos)
        cpp_complex[double] getRadialSolution(BoundaryCondition bc, int dr, int pos)

        double getPolarPoints(int pos)
        double getPolarSolution(int pos)
        double getPolarDerivative(int pos)
        double getPolarSecondDerivative(int pos)
    
    void test_hertz_mode(int j, int m, int k, int n, GeodesicSource& geo)
    void teukolsky_to_hertz_ORG(cpp_complex[double] &Psi, cpp_complex[double] Zteuk, int L, int m, int k, double a, double omega, double lambdaCH)
    void teukolsky_to_hertz_IRG(cpp_complex[double] &Psi, cpp_complex[double] Zteuk, int L, int m, int k, double a, double omega, double lambdaCH)
    void teukolsky_to_hertz_SAAB(cpp_complex[double] &Psi, cpp_complex[double] Zteuk, int L, int m, int k, double a, double omega, double lambdaCH)
    void teukolsky_to_hertz_ASAAB(cpp_complex[double] &Psi, cpp_complex[double] Zteuk, int L, int m, double a, double omega, double lambdaCH)

    void teukolsky_to_hertz_ORG(cpp_complex[double] &PsiIn, cpp_complex[double] &PsiUp, cpp_complex[double] ZteukIn, cpp_complex[double] ZteukUp, int L, int m, int k, double a, double omega, double lambdaCH)
    void teukolsky_to_hertz_IRG(cpp_complex[double] &PsiIn, cpp_complex[double] &PsiUp, cpp_complex[double] ZteukIn, cpp_complex[double] ZteukUp, int L, int m, int k, double a, double omega, double lambdaCH)
    void teukolsky_to_hertz_SAAB(cpp_complex[double] &PsiIn, cpp_complex[double] &PsiUp, cpp_complex[double] ZteukIn, cpp_complex[double] ZteukUp, int L, int m, int k, double a, double omega, double lambdaCH)
    void teukolsky_to_hertz_ASAAB(cpp_complex[double] &PsiIn, cpp_complex[double] &PsiUp, cpp_complex[double] ZteukIn, cpp_complex[double] ZteukUp, int L, int m, double a, double omega, double lambdaCH)

cdef extern from "metriccoeffs.hpp":
    cpp_complex[double] metric_coefficient_ORG(int ai, int bi, int nt, int nr, int nz, int nphi, double a, double r, double z)
    cpp_complex[double] metric_coefficient_IRG(int ai, int bi, int nt, int nr, int nz, int nphi, double a, double r, double z)
    vector[vector[vector[cpp_complex[double]]]] metric_coefficients_ORG_11(double a, vector[double] r, vector[double] z)

cdef dict gauge_dict = {
    "ORG" : Gauge.ORG, 
    "IRG" : Gauge.IRG, 
    "SRG0" : Gauge.SAAB0, 
    "SRG4" : Gauge.SAAB4, 
    "ARG0" : Gauge.ASAAB0, 
    "ARG4" : Gauge.ASAAB4,
    # "SAAB" : Gauge.SAAB, 
    # "ASAAB" : Gauge.ASAAB,
}

cdef Gauge str_to_gauge(unicode gauge_str) except *:
    if gauge_str in gauge_dict.keys():
        return gauge_dict[gauge_str]
    else:
        print("Error")
        TypeError("{} is not a supported gauge.".format(gauge_str))

cdef class TeukolskyMode:
    cdef TeukolskyModeCPP *teukcpp
    cdef int sampleR
    cdef int sampleTh

    def __cinit__(self, int s, int j, int m, int k, int n, KerrGeodesic geo):
        self.teukcpp = new TeukolskyModeCPP(s, j, m, k, n, dereference(geo.geocpp))
        self.sampleR = 1
        self.sampleTh = 1

    def __dealloc__(self):
        del self.teukcpp

    @property
    def radialsamplenumber(self):
        return self.sampleR
    
    @property
    def polarsamplenumber(self):
        return self.sampleTh

    @property
    def spinweight(self):
        return self.teukcpp.getSpinWeight()

    @property
    def spheroidalmode(self):
        return self.teukcpp.getSpheroidalModeNumber()
    
    @property
    def azimuthalmode(self):
        return self.teukcpp.getAzimuthalModeNumber()

    @property
    def radialmode(self):
        return self.teukcpp.getRadialModeNumber()

    @property
    def polarmode(self):
        return self.teukcpp.getPolarModeNumber()

    @property
    def blackholespin(self):
        return self.teukcpp.getBlackHoleSpin()
    
    @property
    def frequency(self):
        return self.teukcpp.getFrequency()

    @property
    def horizonfrequency(self):
        return self.teukcpp.getHorizonFrequency()

    @property
    def eigenvalue(self):
        return self.teukcpp.getEigenvalue()

    @property
    def mincouplingmode(self):
        return self.teukcpp.getMinCouplingModeNumber()
    
    @property
    def maxcouplingmode(self):
        return self.teukcpp.getMaxCouplingModeNumber()

    # some useful aliases
    @property
    def j(self):
        return self.spheroidalmode
    @property
    def m(self):
        return self.azimuthalmode
    @property
    def k(self):
        return self.polarmode
    @property
    def n(self):
        return self.radialmode
    @property
    def omega(self):
        return self.frequency
    @property
    def a(self):
        return self.blackholespin
    @property
    def radialpoints(self):
        return np.array([self.teukcpp.getRadialPoints(i) for i in range(self.sampleR)])
    @property
    def polarpoints(self):
        return np.array([self.teukcpp.getPolarPoints(i) for i in range(self.sampleTh)])
    @property
    def radialsolutions(self):
        return {
                "In":np.array([self.teukcpp.getHomogeneousRadialSolution(BoundaryCondition.In, i) for i in range(self.sampleR)]),
                "Up":np.array([self.teukcpp.getHomogeneousRadialSolution(BoundaryCondition.Up, i) for i in range(self.sampleR)])
        }

    def teukolsky_amplitude(self, unicode bc):
        return self.teukcpp.getTeukolskyAmplitude(str_to_bc(bc))

    def teukolsky_amplitude_precision(self, unicode bc):
        return self.teukcpp.getTeukolskyAmplitudePrecision(str_to_bc(bc))

    def couplingcoefficient(self, int l):
        return self.teukcpp.getCouplingCoefficient(l)
    def radialpoint(self, int i):
        return self.teukcpp.getRadialPoints(i)
    def radialsolution(self, unicode bc, int i):
        return self.teukcpp.getRadialSolution(str_to_bc(bc), i)
    def radialderivative(self, unicode bc, int i):
        return self.teukcpp.getRadialDerivative(str_to_bc(bc), i)
    def radialderivative2(self, unicode bc, int i):
        return self.teukolsky_amplitude(bc)*self.homogeneousradialderivative2(bc, i)
    def homogeneousradialsolution(self, unicode bc, int i):
        return self.teukcpp.getHomogeneousRadialSolution(str_to_bc(bc), i)
    def homogeneousradialderivative(self, unicode bc, int i):
        return self.teukcpp.getHomogeneousRadialDerivative(str_to_bc(bc), i)
    def homogeneousradialderivative2(self, unicode bc, int i):
        return self.teukcpp.getHomogeneousSecondRadialDerivative(str_to_bc(bc), i)
    def polarpoint(self, int i):
        return self.teukcpp.getPolarPoints(i)
    def polarsolution(self, int i):
        return self.teukcpp.getPolarSolution(i)
    def polarderivative(self, int i):
        return self.teukcpp.getPolarDerivative(i)
    def polarderivative2(self, int i):
        return self.teukcpp.getPolarSecondDerivative(i)

    def solve(self, KerrGeodesic geo, unicode method = "AUTO", int nsample = 256, teuk=None, swsh=None):
        self.teukcpp.generateSolutions(dereference(geo.geocpp), str_to_method(method), nsample)
        self.sampleR = self.teukcpp.getRadialSampleNumber()
        self.sampleTh = self.teukcpp.getPolarSampleNumber()

    def flip_spinweight_frequency(self):
        self.teukcpp.flipSpinWeightAndFrequency()
    
    def flip_spinweight(self):
        self.teukcpp.flipSpinWeight()

cdef class HertzMode:
    cdef HertzModeCPP *hertzcpp
    cdef unicode gauge_str
    cdef Gauge gauge_cpp
    cdef int sampleR
    cdef int sampleTh

    def __init__(self, TeukolskyMode teuk, unicode gauge):
        if np.abs(teuk.spinweight) != 2:
            raise ValueError("Hertz mode only accepts Teukolsky solutions with spin-weight -2,+2.")
        self.gauge_cpp = str_to_gauge(gauge)
        self.gauge_str = gauge
        self.hertzcpp = new HertzModeCPP(dereference(teuk.teukcpp), self.gauge_cpp)
        self.sampleR = teuk.sampleR
        self.sampleTh = teuk.sampleTh

    def solve(self):
        self.hertzcpp.generateSolutions()
        
    def __dealloc__(self):
        del self.hertzcpp

    @property
    def radialsamplenumber(self):
        return self.sampleR
    
    @property
    def polarsamplenumber(self):
        return self.sampleTh

    @property
    def gauge(self):
        return self.gauge_str

    @property
    def spinweight(self):
        return self.hertzcpp.getSpinWeight()

    @property
    def spheroidalmode(self):
        return self.hertzcpp.getSpheroidalModeNumber()
    
    @property
    def azimuthalmode(self):
        return self.hertzcpp.getAzimuthalModeNumber()

    @property
    def radialmode(self):
        return self.hertzcpp.getRadialModeNumber()

    @property
    def polarmode(self):
        return self.hertzcpp.getPolarModeNumber()

    @property
    def blackholespin(self):
        return self.hertzcpp.getBlackHoleSpin()
    
    @property
    def frequency(self):
        return self.hertzcpp.getFrequency()

    @property
    def horizonfrequency(self):
        return self.hertzcpp.getHorizonFrequency()

    @property
    def eigenvalue(self):
        return self.hertzcpp.getEigenvalue()

    @property
    def mincouplingmode(self):
        return self.hertzcpp.getMinCouplingModeNumber()
    
    @property
    def maxcouplingmode(self):
        return self.hertzcpp.getMaxCouplingModeNumber()

    @property
    def minscalarcouplingmode(self):
        return self.hertzcpp.getMinScalarCouplingModeNumber()
    
    @property
    def maxscalarcouplingmode(self):
        return self.hertzcpp.getMaxScalarCouplingModeNumber()

    # some useful aliases
    @property
    def j(self):
        return self.spheroidalmode
    @property
    def m(self):
        return self.azimuthalmode
    @property
    def k(self):
        return self.polarmode
    @property
    def n(self):
        return self.radialmode
    @property
    def omega(self):
        return self.frequency
    @property
    def a(self):
        return self.blackholespin

    def hertz_amplitude(self, unicode bc):
        return self.hertzcpp.getHertzAmplitude(str_to_bc(bc))

    def couplingcoefficient(self, int l):
        return self.hertzcpp.getCouplingCoefficient(l)
    def scalarcouplingcoefficient(self, int l):
        return self.hertzcpp.getScalarCouplingCoefficient(l)
    def radialpoint(self, int i):
        return self.hertzcpp.getRadialPoints(i)
    def radialsolution(self, unicode bc, int i):
        return self.hertzcpp.getRadialSolution(str_to_bc(bc), i)
    def radialderivative(self, unicode bc, int i):
        return self.hertzcpp.getRadialDerivative(str_to_bc(bc), i)
    def radialderivative2(self, unicode bc, int i):
        return self.hertz_amplitude(bc)*self.homogeneousradialderivative2(bc, i)
    def homogeneousradialsolution(self, unicode bc, int i):
        return self.hertzcpp.getHomogeneousRadialSolution(str_to_bc(bc), i)
    def homogeneousradialderivative(self, unicode bc, int i):
        return self.hertzcpp.getHomogeneousRadialDerivative(str_to_bc(bc), i)
    def homogeneousradialderivative2(self, unicode bc, int i):
        return self.hertzcpp.getHomogeneousRadialSecondDerivative(str_to_bc(bc), i)
    def polarpoint(self, int i):
        return self.hertzcpp.getPolarPoints(i)
    def polarsolution(self, int i):
        return self.hertzcpp.getPolarSolution(i)
    def polarderivative(self, int i):
        return self.hertzcpp.getPolarDerivative(i)
    def polarderivative2(self, int i):
        return self.hertzcpp.getPolarSecondDerivative(i)

    @property
    def couplingcoefficients(self):
        return np.array([self.couplingcoefficient(i) for i in range(self.mincouplingmode, self.maxcouplingmode+1)])

    @property
    def scalarcouplingcoefficients(self):
        return np.array([self.scalarcouplingcoefficient(i) for i in range(self.minscalarcouplingmode, self.maxscalarcouplingmode+1)])

    @property
    def polarpoints(self):
        return np.array([self.polarpoint(i) for i in range(self.sampleTh)])
    
    @property
    def polarsolutions(self):
        return np.array([self.polarsolution(i) for i in range(self.sampleTh)])
        
    @property
    def polarderivatives(self):
        return np.array([self.polarderivative(i) for i in range(self.sampleTh)])

    @property
    def polarderivatives2(self):
        return np.array([self.polarderivative2(i) for i in range(self.sampleTh)])
    
    @property
    def radialpoints(self):
        return np.array([self.radialpoint(i) for i in range(self.sampleR)])
    @property
    def radialsolutions(self):
        return {
                "In":np.array([self.hertzcpp.getRadialSolution(BoundaryCondition.In, i) for i in range(self.sampleR)]),
                "Up":np.array([self.hertzcpp.getRadialSolution(BoundaryCondition.Up, i) for i in range(self.sampleR)])
        }
    @property
    def radialderivatives(self):
        return {
                "In":np.array([self.hertzcpp.getRadialDerivative(BoundaryCondition.In, i) for i in range(self.sampleR)]),
                "Up":np.array([self.hertzcpp.getRadialDerivative(BoundaryCondition.Up, i) for i in range(self.sampleR)])
        }
    @property
    def radialderivatives2(self):
        return {
                "In":np.array([self.radialderivative2('In', i) for i in range(self.sampleR)]),
                "Up":np.array([self.radialderivative2('Up', i) for i in range(self.sampleR)])
        }
        
def test_hertz_mode_cython(int j, int m, int k, int n, KerrGeodesic geo):
    test_hertz_mode(j, m, k, n, dereference(geo.geocpp))

cdef dict basis_dict = {
    "tetrad": None,
    "coordinate": None
}

def teuk_to_hertz_ORG(cpp_complex[double] ZIn, cpp_complex[double] ZUp, int j, int m, int k, double a, double omega, double lambdaCH):
    cdef cpp_complex[double] PsiIn
    cdef cpp_complex[double] PsiUp
    teukolsky_to_hertz_ORG(PsiIn, PsiUp, ZIn, ZUp, j, m, k, a, omega, lambdaCH)
    return (PsiIn, PsiUp)

def teuk_to_hertz_IRG(cpp_complex[double] ZIn, cpp_complex[double] ZUp, int j, int m, int k, double a, double omega, double lambdaCH):
    cdef cpp_complex[double] PsiIn
    cdef cpp_complex[double] PsiUp
    teukolsky_to_hertz_IRG(PsiIn, PsiUp, ZIn, ZUp, j, m, k, a, omega, lambdaCH)
    return (PsiIn, PsiUp)

def teuk_to_hertz_SRG(cpp_complex[double] ZIn, cpp_complex[double] ZUp, int j, int m, int k, double a, double omega, double lambdaCH):
    cdef cpp_complex[double] PsiIn
    cdef cpp_complex[double] PsiUp
    teukolsky_to_hertz_SAAB(PsiIn, PsiUp, ZIn, ZUp, j, m, k, a, omega, lambdaCH)
    return (PsiIn, PsiUp)

def teuk_to_hertz_ARG(cpp_complex[double] ZIn, cpp_complex[double] ZUp, int j, int m, int k, double a, double omega, double lambdaCH):
    cdef cpp_complex[double] PsiIn
    cdef cpp_complex[double] PsiUp
    teukolsky_to_hertz_ASAAB(PsiIn, PsiUp, ZIn, ZUp, j, m, a, omega, lambdaCH)
    return (PsiIn, PsiUp)

cdef dict metric_component_gauge_dict = {
    "ORG" : {
        (1, 1): None, 
        (1, 3): None,
        (1, 4): None,
        (3, 3): None,
        (4, 4): None}, 
    "IRG" : {
        (2, 2): None,
        (2, 3): None, 
        (2, 4): None,
        (3, 3): None,
        (4, 4): None}, 
    "SRG0" : {
        (2, 2): None, 
        (2, 3): None, 
        (2, 4): None,
        (3, 3): None,
        (4, 4): None}, 
    "SRG4" : {
        (1, 1): None, 
        (1, 3): None, 
        (1, 4): None,
        (3, 3): None,
        (4, 4): None}, 
    "ARG0" : {
        (2, 2): None, 
        (2, 3): None,
        (2, 4): None,
        (3, 3): None,
        (4, 4): None}, 
    "ARG4" : {
        (1, 1): None, 
        (1, 3): None, 
        (1, 4): None,
        (3, 3): None,
        (4, 4): None},
    # "SAAB" : {
    #     (1, 1): None,
    #     (1, 3): None,
    #     (1, 4): None,
    #     (2, 2): None, 
    #     (2, 3): None, 
    #     (2, 4): None,
    #     (3, 3): None,
    #     (4, 4): None}, 
    # "ASAAB" : {
    #     (1, 1): None,
    #     (1, 3): None,
    #     (1, 4): None,
    #     (2, 2): None, 
    #     (2, 3): None, 
    #     (2, 4): None,
    #     (3, 3): None,
    #     (4, 4): None},
}

def metric_11(double a, double r, double z):
    cdef vector[double] rvec = vector[double](1)
    cdef vector[double] zvec = vector[double](1)
    rvec[0] = r
    zvec[0] = z
    cdef cpp_complex[double] temp

    cdef vector[vector[vector[cpp_complex[double]]]] coeffs = metric_coefficients_ORG_11(a, rvec, zvec)
    
    return np.array(coeffs).squeeze()

def metric_coefficient_S4(int alpha, int beta, int nt, int nr, int nz, int np, double a, double r, double z):
    return metric_coefficient_ORG(alpha, beta, nt, nr, nz, np, a, r, z)

def metric_coefficient_S0(int alpha, int beta, int nt, int nr, int nz, int np, double a, double r, double z):
    return metric_coefficient_IRG(alpha, beta, nt, nr, nz, np, a, r, z)

cdef class MetricModeGenerator:
    cdef unicode gauge_str
    cdef Gauge gauge_cpp
    cdef unicode basis

    def __init__(self, unicode gauge, unicode basis="tetrad"):
        self.gauge_str = gauge
        self.gauge_cpp = str_to_gauge(gauge)
        if basis in basis_dict.keys():
            self.basis = basis
        else:
            raise TypeError('{} is not a valid basis. Must be tetrad or coordinate.'.format(basis))
    
    def __call__(self, HertzMode hertz, int ai, int bi):
        if self.basis == "tetrad":
            return self.tetradcomponent(hertz, ai, bi)
        else:
            return self.tetradcomponent(hertz, ai, bi)
    
    def tetradcomponent(self, HertzMode hertz, int ai, int bi):
        if hertz.gauge is not self.gauge_str:
            raise TypeError("Hertz potential in {} gauge. Must be in {} gauge".format(hertz.gauge, self.gauge_str))
        cdef cpp_complex[double] habIn = cpp_complex[double](0., 0.)
        cdef cpp_complex[double] habUp = cpp_complex[double](0., 0.)
        cdef cpp_complex[double] habbase, dPsiIn, dPsiUp, dS
        cdef int atemp
        cdef double a, r, z
        cdef cpp_complex[double] im = cpp_complex[double](0., hertz.azimuthalmode)
        cdef cpp_complex[double] iomega = cpp_complex[double](0., hertz.frequency)
        a = hertz.blackholespin
        r = hertz.radialpoint(0)
        z = np.cos(hertz.polarpoint(0))
        if(ai < bi):
            atemp = ai
            ai = bi
            bi = atemp
        if (ai, bi) in metric_component_gauge_dict[self.gauge_str].keys():
            for nt in range(3):
                for nr in range(3):
                    if nr == 0:
                        dPsiIn = hertz.radialsolution("In", 0)
                        dPsiUp = hertz.radialsolution("Up", 0)
                    elif nr == 1:
                        dPsiIn = hertz.radialderivative("In", 0)
                        dPsiUp = hertz.radialderivative("Up", 0)
                    elif nr == 2:
                        dPsiIn = hertz.radialderivative2("In", 0)
                        dPsiUp = hertz.radialderivative2("Up", 0)
                    else:
                        dPsiIn = cpp_complex[double](0., 0.)
                        dPsiUp = cpp_complex[double](0., 0.)
                    for nz in range(3):
                        if nz == 0:
                            dS = hertz.polarsolution(0)
                        elif nz == 1: # SWSH are functions of theta but the metric coefficients are organized wrt to d/dz
                            dS = -hertz.polarderivative(0)/np.sqrt(1. - z**2)
                        elif nz == 2:
                            dS = hertz.polarderivative2(0)/(1. - z**2) - z*hertz.polarderivative(0)/(1. - z**2)**(1.5)
                        else:
                            dS = cpp_complex[double](0., 0.)
                        for nph in range(3):
                            if nt + nr + nz + nph <= 2:
                                habbase = metric_coefficient_S0(ai, bi, nt, nr, nz, nph, a, r, z)*dS*(im)**nph*(-1.*iomega)**nt
                                habIn += habbase*dPsiIn
                                habUp += habbase*dPsiUp
            return {"In": habIn, "Up": habUp}
        else:
            return {"In": habIn, "Up": habUp}



from cybhpt_full import FluxList as FluxListCython
from cybhpt_full import flux as FluxCython

class FluxList:
    """A class for storing fluxes of energy, angular momentum, and Carter constant.

    Parameters
    ----------
    fluxes : list of dicts, optional
        A list containing three dictionaries, each with keys "I" and "H" representing the fluxes at infinity and on the horizon, respectively.
        If not provided, initializes with zero fluxes for all three components.

    Attributes
    ----------
    fluxes : list of dicts
        A list containing three dictionaries, each with keys "I" and "H" representing the fluxes at infinity and on the horizon, respectively.
        The first dictionary corresponds to energy flux,
        the second to angular momentum flux, and the third to Carter constant flux.

    Methods
    -------
    __call__():
        Returns the list of fluxes.
    """
    def __init__(self, fluxes=None):
        if fluxes is None:
            self.fluxes = [{"I": 0., "H": 0.}, {"I": 0., "H": 0.}, {"I": 0., "H": 0.}]
        elif len(fluxes) != 3:
            self.fluxes = [{"I": 0., "H": 0.}, {"I": 0., "H": 0.}, {"I": 0., "H": 0.}]
        else:
            self.fluxes = fluxes

    # def __add__(self, fluxlist2):
    #     for i in range(3):
    #         for bc in ["H", "I"]:
    #             self.fluxes[i][bc] += fluxlist2.fluxes[i][bc]
    
    @property
    def energy(self):
        return self.fluxes[0]

    @property
    def angularmomentum(self):
        return self.fluxes[1]

    @property
    def carterconstant(self):
        return self.fluxes[2]
    
    def __call__(self):
        return self.fluxes

    
class FluxMode:
    """A class for computing fluxes of energy, angular momentum, and Carter constant for a given Teukolsky mode.

    Parameters
    ----------
    geo : KerrGeodesic class instance
        KerrGeodesic object containing the background motion of the point-particle source.

    teuk : TeukolskyMode object
        TeukolskyMode object containing mode solutions to the point-particle-sourced Teukolsky equation.

    Attributes
    ----------
    energy : dict
        A dictionary containing the energy flux at infinity and on the horizon.
        The keys are "I" for infinity and "H" for the horizon.
    angularmomentum : dict
        A dictionary containing the angular momentum flux at infinity and on the horizon.
        The keys are "I" for infinity and "H" for the horizon.
    carterconstant : dict
        A dictionary containing the Carter constant flux at infinity and on the horizon.
        The keys are "I" for infinity and "H" for the horizon.
    fluxes : FluxList
        A FluxList object containing the energy, angular momentum, and Carter constant fluxes.
    horizonfluxes : list
        A list containing the energy, angular momentum, and Carter constant fluxes on the horizon.
    infinityfluxes : list
        A list containing the energy, angular momentum, and Carter constant fluxes at infinity.
    totalfluxes : list
        A list containing the total fluxes, which are the sum of the horizon and infinity fluxes.
        """
    def __init__(self, geo, teuk):
        self.base = FluxCython(teuk.spinweight, geo.base, teuk.base)

    @property
    def energy(self):
        return self.base.energy

    @property
    def angularmomentum(self):
        return self.base.angularmomentum

    @property
    def carterconstant(self):
        return self.base.carterconstant
    
    @property
    def Edot(self):
        return self.energy
    @property
    def Ldot(self):
        return self.angularmomentum
    @property
    def Qdot(self):
        return self.carterconstant
    
    @property
    def fluxes(self):
        return FluxList([self.energy, self.angularmomentum, self.carterconstant])

    @property
    def horizonfluxes(self):
        return [self.energy["H"], self.angularmomentum["H"], self.carterconstant["H"]]
    
    @property
    def infinityfluxes(self):
        return [self.energy["I"], self.angularmomentum["I"], self.carterconstant["I"]]
    
    @property
    def totalfluxes(self):
        return [self.energy["I"] + self.energy["H"], self.angularmomentum["I"] + self.angularmomentum["H"], self.carterconstant["I"] + self.carterconstant["H"]]



from libcpp.vector cimport vector
from libcpp.complex cimport complex as cpp_complex
from cython.operator cimport dereference
from libcpp.string cimport string as cpp_string
import numpy as np
cimport numpy as np

include "teukolsky_wrap.pyx"

cdef extern from "fluxes.hpp": 
    cdef cppclass FluxesCPP "Fluxes":
        FluxesCPP()
        FluxesCPP(double inf, double hor)
        double infinity
        double horizon

    cdef cppclass FluxListCPP "FluxList":
        FluxListCPP()
        FluxesCPP Edot
        FluxesCPP Ldot
        FluxesCPP Qdot

    FluxListCPP flux_mode(int s, GeodesicSource& geo, TeukolskyModeCPP& teukMode, int include_minus_m)

    FluxListCPP fluxes(GeodesicSource& geo)
    FluxListCPP fluxes(int s, GeodesicSource& geo)
    FluxListCPP flux_l(int s, int L, GeodesicSource& geo)
    FluxListCPP flux_lm(int s, int L, int m, GeodesicSource& geo)
    FluxListCPP flux_lm_sum(int s, int L, int m, GeodesicSource& geo)
    FluxListCPP flux_lmk(int s, int L, int m, int k, GeodesicSource& geo)
    FluxListCPP flux_lmk_sum(int s, int L, int m, int k, GeodesicSource& geo)

    void full_flux_parallel_l(int s, GeodesicSource geo, int modeMax, cpp_string dir)
    void full_flux_parallel_lm(GeodesicSource geo, int lMax, cpp_string dir)

cdef class FluxList:
    cdef FluxListCPP *fluxlistcpp

    def __cinit__(self):
        self.fluxlistcpp = new FluxListCPP()

    def __dealloc__(self):
        del self.fluxlistcpp

    def zero_fluxes(self):
        self.fluxlistcpp.Edot.infinity = 0.
        self.fluxlistcpp.Ldot.infinity = 0.
        self.fluxlistcpp.Qdot.infinity = 0.

        self.fluxlistcpp.Edot.horizon = 0.
        self.fluxlistcpp.Ldot.horizon = 0.
        self.fluxlistcpp.Qdot.horizon = 0.

    cdef set_fluxes(self, FluxListCPP fluxes):
        self.fluxlistcpp.Edot.infinity = fluxes.Edot.infinity
        self.fluxlistcpp.Ldot.infinity = fluxes.Ldot.infinity
        self.fluxlistcpp.Qdot.infinity = fluxes.Qdot.infinity

        self.fluxlistcpp.Edot.horizon = fluxes.Edot.horizon
        self.fluxlistcpp.Ldot.horizon = fluxes.Ldot.horizon
        self.fluxlistcpp.Qdot.horizon = fluxes.Qdot.horizon

    def set_infinity_fluxes(self, double Edot, double Ldot, double Qdot):
        self.fluxlistcpp.Edot.infinity = Edot
        self.fluxlistcpp.Ldot.infinity = Ldot
        self.fluxlistcpp.Qdot.infinity = Qdot

    def set_horizon_fluxes(self, double Edot, double Ldot, double Qdot):
        self.fluxlistcpp.Edot.horizon = Edot
        self.fluxlistcpp.Ldot.horizon = Ldot
        self.fluxlistcpp.Qdot.horizon = Qdot

    def add_infinity_fluxes(self, double Edot, double Ldot, double Qdot):
        self.fluxlistcpp.Edot.infinity += Edot
        self.fluxlistcpp.Ldot.infinity += Ldot
        self.fluxlistcpp.Qdot.infinity += Qdot

    def add_horizon_fluxes(self, double Edot, double Ldot, double Qdot):
        self.fluxlistcpp.Edot.horizon += Edot
        self.fluxlistcpp.Ldot.horizon += Ldot
        self.fluxlistcpp.Qdot.horizon += Qdot

    def set_infinity_fluxes(self, double Edot, double Ldot, double Qdot):
        self.fluxlistcpp.Edot.infinity += Edot
        self.fluxlistcpp.Ldot.infinity += Ldot
        self.fluxlistcpp.Qdot.infinity += Qdot

    def add_fluxes(self, double EdotH, double LdotH, double QdotH, double EdotI, double LdotI, double QdotI):
        self.fluxlistcpp.Edot.horizon += EdotH
        self.fluxlistcpp.Ldot.horizon += LdotH
        self.fluxlistcpp.Qdot.horizon += QdotH

        self.fluxlistcpp.Edot.infinity += EdotI
        self.fluxlistcpp.Ldot.infinity += LdotI
        self.fluxlistcpp.Qdot.infinity += QdotI

    @property
    def energy(self):
        return {
            "I": self.fluxlistcpp.Edot.infinity,
            "H": self.fluxlistcpp.Edot.horizon
        }

    @property
    def angularmomentum(self):
        return {
            "I": self.fluxlistcpp.Ldot.infinity,
            "H": self.fluxlistcpp.Ldot.horizon
        }

    @property
    def carterconstant(self):
        return {
            "I": self.fluxlistcpp.Qdot.infinity,
            "H": self.fluxlistcpp.Qdot.horizon
        }


def flux(int s, KerrGeodesic geo, TeukolskyMode teuk):
    cdef FluxListCPP fluxescpp = flux_mode(s, dereference(geo.geocpp), dereference(teuk.teukcpp), include_minus_m = 0)
    fluxes = FluxList()
    fluxes.set_fluxes(fluxescpp)
    return fluxes

# def full_flux_parallel_l_py(int s, KerrGeodesic geo, int modeMax, unicode wdir):
#     full_flux_parallel_l(s, dereference(geo.geocpp), modeMax, wdir.encode())

# def full_flux_parallel_lm_py(KerrGeodesic geo, int lmax, unicode wdir):
#     full_flux_parallel_lm(dereference(geo.geocpp), lmax, wdir.encode())


from libcpp.vector cimport vector
from libcpp.complex cimport complex as cpp_complex
import numpy as np
cimport numpy as np

cdef extern from "gsl/gsl_errno.h":
    void gsl_set_error_handler_off()

# call at import-time
gsl_set_error_handler_off()

cdef extern from "radialsolver.hpp":
    cdef enum SolutionMethod:
        AUTO, MST, ASYM, HBL, GSN, TEUK

    cdef enum BoundaryCondition:
        In, Up

    cdef cppclass Result:
        Result(cpp_complex[double], cpp_complex[double])
        cpp_complex[double] getValue()
        cpp_complex[double] getAccuracy()


    cdef cppclass RadialTeukolskyCPP "RadialTeukolsky":
        RadialTeukolskyCPP(double a, int s, int l, int m, double omega, vector[double] r) except +
        # RadialTeukolsky(double a, int s, int l, int m, double omega, double la, vector[double] r)  except +

        double getBlackHoleSpin()
        int getSpinWeight()
        int getSpheroidalModeNumber()
        int getAzimuthalModeNumber()
        double getModeFrequency()
        double getSpinWeightedSpheroidalEigenvalue()

        void generateRetardedBoundaryConditions(SolutionMethod method)
        void setBoundaryConditions(BoundaryCondition bc, cpp_complex[double] R, cpp_complex[double] Rp, double r)
        void generateSolutions(SolutionMethod method)
        void generateSolutions(BoundaryCondition bc, SolutionMethod method)
        int resampleSolutions(vector[double] radialSamples)
        void flipSpinWeight()

        vector[double] getRadialPoints()
        double getRadialPoints(int pos)
        double getBoundaryPoint(BoundaryCondition bc)
        Result getBoundarySolution(BoundaryCondition bc)
        Result getBoundaryDerivative(BoundaryCondition bc)

        vector[cpp_complex[double]] getSolution(BoundaryCondition bc)
        vector[cpp_complex[double]] getDerivative(BoundaryCondition bc)
        vector[cpp_complex[double]] getSecondDerivative(BoundaryCondition bc)
        cpp_complex[double] getSolution(BoundaryCondition bc, int pos)
        cpp_complex[double] getDerivative(BoundaryCondition bc, int pos)
        cpp_complex[double] getSecondDerivative(BoundaryCondition bc, int pos)

    void flip_spin_of_radial_teukolsky_TS(cpp_complex[double] &RinFlip, cpp_complex[double] &RinPFlip, BoundaryCondition bc, int s, int m, double a, double omega, double la, double r, cpp_complex[double] Rin, cpp_complex[double] RinP)
    double teukolsky_starobinsky_constant(int s, int m, double a, double omega, double lambdaCH)
    double teukolsky_starobinsky_constant_D(int m, double a, double omega, double lambdaCH)
    cpp_complex[double] teukolsky_starobinsky_complex_constant(int j, int m, double a, double omega, double lambdaCH)
    cpp_complex[double] teukolsky_starobinsky_amplitude(BoundaryCondition bc, int s, int m, double a, double omega, double lambdaCH)

cdef extern from "monodromy.hpp":
    cpp_complex[double] nu_solver_monodromy(int s, int l, int m, double q, double eps, double la)

cdef extern from "nusolver.hpp":
    cpp_complex[double] nu_solver(double q, int s, int l, int m, double epsilon)

cdef extern from "hypergeo_f.hpp":
    cpp_complex[double] hypergeo_2F1_complex(cpp_complex[double] a, cpp_complex[double] b, cpp_complex[double] c, cpp_complex[double] x)

def hypergeo_2F1(cpp_complex[double] a, cpp_complex[double] b, cpp_complex[double] c, cpp_complex[double] x):
    return hypergeo_2F1_complex(a, b, c, x)

def renormalized_angular_momentum(int s, int l, int m, double a, double omega):
    return nu_solver(a, s, l, m, 2.*omega)

def renormalized_angular_momentum_monodromy(int s, int l, int m, double a, double omega, double la):
    return nu_solver_monodromy(s, l, m, a, 2.*omega, la)

cdef dict bc_dict = {
    "In" : BoundaryCondition.In,
    "Up" : BoundaryCondition.Up
}

cdef dict method_dict = {
    "AUTO" : SolutionMethod.AUTO, 
    "MST" : SolutionMethod.MST, 
    "ASYMP" : SolutionMethod.ASYM, 
    "HBL" : SolutionMethod.HBL, 
    "GSN" : SolutionMethod.GSN, 
    "TEUK" : SolutionMethod.TEUK
}

def available_methods():
    """
    Returns a list of available solution methods.
    """
    return list(method_dict.keys())

def flip_spin_of_solutions(unicode bc, int s, double a, int m, double omega, double la, double r, cpp_complex[double] R, cpp_complex[double] Rp):
    cdef cpp_complex[double] R0, RP0
    flip_spin_of_radial_teukolsky_TS(R0, RP0, str_to_bc(bc), s, m, a, omega, la, r, R, Rp)
    return (R0, RP0)

cdef BoundaryCondition str_to_bc(unicode bc_str):
    if bc_str in bc_dict.keys():
        return bc_dict[bc_str]
    else:
        raise ValueError("{} is not a supported boundary condition.".format(bc_str))

cdef SolutionMethod str_to_method(unicode method_str):
    if method_str in method_dict.keys():
        return method_dict[method_str]
    else:
        raise ValueError("{} is not a supported solution method.".format(method_str))

def teukolsky_starobinsky_transformation_amplitude(unicode bc, int s, int m, double a, double omega, double lambdaCH):
    return teukolsky_starobinsky_amplitude(str_to_bc(bc), s, m, a, omega, lambdaCH)

def teukolsky_starobinsky_const(int j, int m, double a, double omega, double lambdaCH):
    return teukolsky_starobinsky_complex_constant(j, m, a, omega, lambdaCH)

def teukolsky_starobinsky_const_squared(int s, int m, double a, double omega, double lambdaCH):
    return teukolsky_starobinsky_constant(s, m, a, omega, lambdaCH)

cdef class RadialTeukolsky:
    cdef RadialTeukolskyCPP *teukcpp

    def __cinit__(self, double a, int s, int l, int m, double omega, np.ndarray[ndim=1, dtype=np.float64_t] r not None):
        cdef int length_r = r.shape[0]
        cdef vector[double] rvec
        rvec = vector[double](length_r)
        rvec.assign(&r[0], &r[0] + length_r)
        self.teukcpp = new RadialTeukolskyCPP(a, s, l, m, omega, rvec)
        if self.teukcpp == NULL:
            raise MemoryError('Not enough memory.')
    
    def __dealloc__(self):
        del self.teukcpp
    
    @property
    def blackholespin(self):
        return self.teukcpp.getBlackHoleSpin()
    
    @property
    def spinweight(self):
        return self.teukcpp.getSpinWeight()
    @property
    def s(self):
        return self.spinweight

    @property
    def spheroidalmode(self):
        return self.teukcpp.getSpheroidalModeNumber()

    @property
    def j(self):
        return self.spheroidalmode

    @property
    def azimuthalmode(self):
        return self.teukcpp.getAzimuthalModeNumber()
    
    @property
    def m(self):
        return self.azimuthalmode

    @property
    def frequency(self):
        return self.teukcpp.getModeFrequency()
    
    @property
    def mode_frequency(self):
        return self.frequency

    @property
    def omega(self):
        return self.frequency

    @property
    def eigenvalue(self):
        return self.teukcpp.getSpinWeightedSpheroidalEigenvalue()
    
    def solve_bc(self, unicode method):
        self.teukcpp.generateRetardedBoundaryConditions(str_to_method(method))

    def set_bc(self, unicode bc, cpp_complex[double] R, cpp_complex[double] Rp, double r):
        self.teukcpp.setBoundaryConditions(str_to_bc(bc), R, Rp, r)

    def solve(self, unicode method="AUTO", unicode bc="None"):
        if bc == "None":
            self.teukcpp.generateSolutions(str_to_method(method))
        else:
            self.teukcpp.generateSolutions(str_to_bc(bc), str_to_method(method))
    
    def flip_spinweight(self):
        self.teukcpp.flipSpinWeight()

    def radialpoint(self, int pos):
        return self.teukcpp.getRadialPoints(pos)

    def boundarypoint(self, unicode bc):
        return self.teukcpp.getBoundaryPoint(str_to_bc(bc))

    def boundarysolution(self, unicode bc):
        return self.teukcpp.getBoundarySolution(str_to_bc(bc)).getValue()
    
    def boundaryderivative(self, unicode bc):
        return self.teukcpp.getBoundaryDerivative(str_to_bc(bc)).getValue()

    def solution(self, unicode bc, int pos):
        return self.teukcpp.getSolution(str_to_bc(bc), pos)

    def derivative(self, unicode bc, int pos):
        return self.teukcpp.getDerivative(str_to_bc(bc), pos)

    def second_derivative(self, unicode bc, int pos):
        return self.teukcpp.getSecondDerivative(str_to_bc(bc), pos)
    
    def derivative2(self, unicode bc, int pos):
        return self.teukcpp.getSecondDerivative(str_to_bc(bc), pos)


from cybhpt_full import TeukolskyMode as TeukolskyModeCython

class TeukolskyMode:
    """A class for computing Teukolsky modes sourced by a point-particle orbiting in a Kerr background.
    
    Parameters
    ----------
    s : int
        The spin weight of the Teukolsky mode.
    j : int
        The spheroidal harmonic mode number.
    m : int
        The azimuthal harmonic mode number.
    k : int
        The polar harmonic mode number.
    n : int
        The radial harmonic mode number.
    geo : KerrGeodesic class instance
        KerrGeodesic object containing the background motion of the point-particle source.
    auto_solve : bool, optional
        If True, the Teukolsky equation is automatically solved upon initialization. Default is False

    Attributes
    ----------
    spinweight : int
        The spin weight of the Teukolsky mode.
    spheroidalmode : int
        The spheroidal harmonic mode number.
    azimuthalmode : int
        The azimuthal harmonic mode number.
    radialmode : int
        The radial harmonic mode number.
    polarmode : int
        The polar harmonic mode number.
    blackholespin : float
        The spin of the black hole in the Kerr background.
    frequency : float
        The frequency of the Teukolsky mode.
    horizonfrequency : float
        The frequency of the mode at the horizon.
    eigenvalue : float
        The spheroidal eigenvalue of the Teukolsky mode.
    mincouplingmode : int
        The minimum l-mode used for coupling the spherical and spheroidal harmonics
    maxcouplingmode : int
        The maximum l-mode used for coupling the spherical and spheroidal harmonics
    j : int
        Alias for spheroidalmode.
    m : int
        Alias for azimuthalmode.
    k : int
        Alias for polarmode.
    n : int
        Alias for radialmode.
    omega : float
        Alias for frequency.
    a : float
        Alias for blackholespin.
    
    Methods
    -------
    solve(geo, method = "AUTO", nsamples = 256, teuk = None, swsh = None)
        Solve the Teukolsky equation for the given mode and geodesic.
    flipspinweight()
        Flip the spin weight of the Teukolsky mode.
    flipspinweightandfrequency()
        Flip the spin weight and frequency of the Teukolsky mode.
    couplingcoefficient(l)
        Compute the coupling coefficient for the given l-mode.
    radialpoint(pos)
        Compute the radial point for the given position.
    radialsolution(bc, pos)
        Compute the radial solution for the given boundary condition and position.
    radialderivative(bc, pos)
        Compute the radial derivative for the given boundary condition and position.
    radialderivative2(bc, pos)
        Compute the second radial derivative for the given boundary condition and position.
    homogeneousradialsolution(bc, pos)
        Compute the homogeneous radial solution for the given boundary condition and position.
    homogeneousradialderivative(bc, pos)
        Compute the homogeneous radial derivative for the given boundary condition and position.
    homogeneousradialderivative2(bc, pos)
        Compute the second homogeneous radial derivative for the given boundary condition and position.
    polarpoint(pos)
        Compute the polar point for the given position.
    polarsolution(pos)
        Compute the polar solution for the given position.
    polarderivative(pos)
        Compute the polar derivative for the given position.
    polarderivative2(pos)
        Compute the second polar derivative for the given position.
    amplitude(bc)
        Compute the Teukolsky amplitude for the given boundary condition.
    precision(bc)
        Compute the precision of the Teukolsky amplitude for the given boundary condition.
    """
    def __init__(self, s, j, m, k, n, geo, auto_solve = False):
        self.base = TeukolskyModeCython(s, j, m, k, n, geo.base)
        if auto_solve:
            self.solve(geo)

    @property
    def spinweight(self):
        return self.base.spinweight

    @property
    def spheroidalmode(self):
        return self.base.spheroidalmode
    
    @property
    def azimuthalmode(self):
        return self.base.azimuthalmode

    @property
    def radialmode(self):
        return self.base.radialmode

    @property
    def polarmode(self):
        return self.base.polarmode

    @property
    def blackholespin(self):
        return self.base.blackholespin
    
    @property
    def frequency(self):
        return self.base.frequency

    @property
    def horizonfrequency(self):
        return self.base.horizonfrequency

    @property
    def eigenvalue(self):
        return self.base.eigenvalue

    @property
    def mincouplingmode(self):
        return self.base.mincouplingmode
    
    @property
    def maxcouplingmode(self):
        return self.base.maxcouplingmode

    # some useful aliases
    @property
    def j(self):
        return self.spheroidalmode
    @property
    def m(self):
        return self.azimuthalmode
    @property
    def k(self):
        return self.polarmode
    @property
    def n(self):
        return self.radialmode
    @property
    def omega(self):
        return self.frequency
    @property
    def a(self):
        return self.blackholespin
    
    def solve(self, geo, method = "AUTO", nsamples = 256, teuk = None, swsh = None):
        """Solve the Teukolsky equation for the given mode and geodesic.
        Parameters
        ----------
        geo : KerrGeodesic class instance
            KerrGeodesic object containing the background motion of the point-particle source.
        method : str, optional
            The method to use for solving the Teukolsky equation. Default is "AUTO".
        nsamples : int, optional
            The number of samples to use for the solution. Default is 256.
        teuk : RadialTeukolsky, optional
            RadialTeukolsky object to use for constructing the radial Green function. Default is None.
        swsh : SpheroidalHarmonicMode, optional
            SpheroidalHarmonic object to use for coupling with spheroidal harmonics. Default is None.

        """
        if teuk is None or swsh is None:
            self.base.solve(geo.base, method, nsamples)
        else:
            self.base.solve(geo.base, method, nsamples, teuk.base, swsh.base)

    """
    Flips the spin-weight of the Teukolsky solutions from :math:`s \rightarrow -s`
    """
    def flipspinweight(self):
        self.base.flip_spinweight()

    """
    Flips the spin-weight and frequency of the Teukolsky solutions from :math:`s \rightarrow -s` and :math:`\omega \rightarrow -\omega`
    """
    def flipspinweightandfrequency(self):
        self.base.flip_spinweight_frequency()

    """
    Spherical-spheroidal mixing coefficient between a spherical harmonic :math:`l` mode with a spheroidal :math:`j` mode

    :param l: spherical harmonic mode
    :type l: int
    """
    def couplingcoefficient(self, l):
        return self.base.couplingcoefficient(l)

    def radialpoint(self, pos):
        return self.base.radialpoint(pos)
    
    def radialsolution(self, bc, pos):
        return self.base.radialsolution(bc, pos)
    
    def radialderivative(self, bc, pos):
        return self.base.radialderivative(bc, pos)
    
    def radialderivative2(self, bc, pos):
        return self.base.radialderivative2(bc, pos)
    
    def homogeneousradialsolution(self, bc, pos):
        return self.base.homogeneousradialsolution(bc, pos)
    
    def homogeneousradialderivative(self, bc, pos):
        return self.base.homogeneousradialderivative(bc, pos)
    
    def homogeneousradialderivative2(self, bc, pos):
        return self.base.homogeneousradialderivative2(bc, pos)
    
    def polarpoint(self, pos):
        return self.base.polarpoint(pos)
    
    def polarsolution(self, pos):
        return self.base.polarsolution(pos)
    
    def polarderivative(self, pos):
        return self.base.polarderivative(pos)
    
    def polarderivative2(self, pos):
        return self.base.polarderivative2(pos)
    
    def amplitude(self, bc):
        return self.base.teukolsky_amplitude(bc)
    
    def precision(self, bc):
        return self.base.teukolsky_amplitude_precision(bc)


from cybhpt_full import RadialTeukolsky as RadialTeukolskyCython
from cybhpt_full import available_methods as available_methods_cython
from cybhpt_full import renormalized_angular_momentum as nu_cython
from cybhpt_full import renormalized_angular_momentum_monodromy as nu_2_cython
from cybhpt_full import hypergeo_2F1 as hypergeo_2F1_cython
import numpy as np

def available_methods():
    """
    Returns a list of available solution methods.
    """
    return available_methods_cython()

def renormalized_angular_momentum(s, l, m, a, omega):
    return nu_cython(s, l, m, a, omega)

def renormalized_angular_momentum_monodromy(s, l, m, a, omega, la):
    return nu_2_cython(s, l, m, a, omega, la)

def hypergeo_2F1(a, b, c, x):
    return hypergeo_2F1_cython(a, b, c, x)

class RadialTeukolsky:
    """A class for solving the homogeneous radial Teukolsky equation.

    Parameters
    ----------
    s : int
        The spin weight of the field
    j : int
        The spheroidal harmonic mode number
    m : int
        The azimuthal harmonic mode number
    a : float
        The black hole spin parameter
    omega : float
        The frequency of the mode
    r : numpy.ndarray
        A numpy array of radial points at which to evaluate the solution

    Attributes
    ----------
    radialpoints : numpy.ndarray
        A numpy array of radial points at which the solution is evaluated.
    base : RadialTeukolskyCython
        The underlying Cython object that performs the computations.
    nsamples : int
        The number of radial points in the radialpoints array.
    
    Properties
    ----------
    blackholespin : float
        The black hole spin parameter.
    spinweight : int
        The spin weight of the field.
    s : int
        Alias for spinweight.
    spheroidalmode : int
        The spheroidal harmonic mode number.
    j : int
        Alias for spheroidalmode.
    azimuthalmode : int
        The azimuthal harmonic mode number.
    m : int
        Alias for azimuthalmode.
    frequency : float
        The frequency of the mode.
    mode_frequency : float
        Alias for frequency.
    omega : float
        Alias for frequency.
    eigenvalue : float
        The spheroidal eigenvalue of the radial Teukolsky equation.
    
    Methods
    -------
    solveboundarycondition(method)
        Solves the boundary condition for the radial Teukolsky equation.
    setboundarycondition(bc, R, Rp, r)
        Sets the boundary condition for the radial Teukolsky equation.
    solve(method="AUTO", bc=None)
        Solves the radial Teukolsky equation.
    flipspinweight()
        Flips the spin weight of the field.
    radialpoint(pos)
        Returns the radial point at the given position.
    boundarypoint(bc)
        Returns the boundary point for the given boundary condition.
    boundarysolution(bc)
        Returns the solution at the boundary for the given boundary condition.
    boundaryderivative(bc)
        Returns the derivative at the boundary for the given boundary condition.
    radialsolution(bc, pos)
        Returns the solution at the radial point for the given boundary condition and position.
    radialderivative(bc, pos)
        Returns the derivative at the radial point for the given boundary condition and position.
    radialderivative2(bc, pos)
        Returns the second derivative at the radial point for the given boundary condition and position.
    radialsolutions(bc)
        Returns the solutions at all radial points for the given boundary condition.
    radialderivatives(bc)
        Returns the derivatives at all radial points for the given boundary condition.
    radialderivatives2(bc)
        Returns the second derivatives at all radial points for the given boundary condition.
    __call__(bc, deriv=0)
        Returns the solutions, first derivatives, or second derivatives at all radial points for the given boundary condition.
        The `deriv` parameter specifies which derivative to return: 0 for solutions,
        1 for first derivatives, and 2 for second derivatives. If `deriv` is not 0, 1, or 2, a ValueError is raised.
    """
    def __init__(self, s, j, m, a, omega, r):
        if a < 0 or a > 1:
            raise ValueError(f"Black hole spin parameter {a} must be in the range [0, 1].")
        if j < np.abs(m):
            raise ValueError(f"Spheroidal harmonic mode number {j} must be greater than or equal to the absolute value of azimuthal harmonic mode number {m}.")
        if np.any(r <= 1 + np.sqrt(1 - a**2)):
            raise ValueError(f"Radial point {r} must be greater than horizon radius r_+ = {1 + np.sqrt(1 - a**2)}.")
        if isinstance(r, list) or (isinstance(r, np.ndarray) and r.ndim > 0):
            self.radialpoints = np.asarray(r)
            self.nsamples = self.radialpoints.shape[0]
        else:
            raise AttributeError("Radial points must be a list or a numpy array.")
        
        if self.nsamples == 0:
            raise ValueError("Radial points array is empty.")
        self.base = RadialTeukolskyCython(a, s, j, m, omega, self.radialpoints)


    @property
    def blackholespin(self):
        return self.base.blackholespin
    
    @property
    def spinweight(self):
        return self.base.spinweight
    
    @property
    def s(self):
        return self.spinweight

    @property
    def spheroidalmode(self):
        return self.base.spheroidalmode

    @property
    def j(self):
        return self.spheroidalmode

    @property
    def azimuthalmode(self):
        return self.base.azimuthalmode
    
    @property
    def m(self):
        return self.azimuthalmode

    @property
    def frequency(self):
        return self.base.frequency
    
    @property
    def mode_frequency(self):
        return self.frequency

    @property
    def omega(self):
        return self.frequency

    @property
    def eigenvalue(self):
        return self.base.eigenvalue
    
    def solveboundarycondition(self, method):
        """Solves the boundary condition for the radial Teukolsky equation.

        Parameters
        ----------
        method : str
            The method to use for solving the boundary condition. Default is "AUTO".
        """
        self.base.solve_bc(method)

    def setboundarycondition(self, bc, R, Rp, r):
        """Sets the boundary condition for the radial Teukolsky equation.

        Parameters
        ----------
        bc : str
            The boundary condition to set. Can be "In" for horizon or "Up" for infinity.
        R : float
            The boundary condition function value at the radial point.
        Rp : float
            The derivative of the boundary condition function at the radial point.
        r : float
            The radial point at which the boundary condition is defined.
        """
        self.base.set_bc(bc, R, Rp, r)

    def solve(self, method = "AUTO", bc = None):
        """Solves the radial Teukolsky equation.

        Parameters
        ----------
        method : str, optional
            The method to use for solving the equation. Default is "AUTO".
        bc : str, optional
            Specifies which homogeneous solutions to compute. If None, both "In" (horizon) and "Up" (infinity) solutions are computed.
            If "In", only the horizon solution is computed. If "Up", only the infinity solution is computed.
        """
        if bc is None:
            self.base.solve(method, "None")
        else:
            self.base.solve(method, bc)

    def flipspinweight(self):
        """Flips the sign of the spin weight of the field."""
        self.base.flip_spinweight()

    def radialpoint(self, pos):
        """Returns the radial point at the given position.
        Parameters
        ----------
        pos : float
            The position at which to evaluate the radial point.
        Returns
        -------
        float
            The radial point at the given position.
        """
        return self.base.radialpoint(pos)
    
    def boundarypoint(self, bc):
        """
        Returns the boundary point for the given boundary condition.
        Parameters
        ----------
        bc : str
            The boundary condition to evaluate. Can be "In" for horizon or "Up" for infinity.
        Returns
        -------
        float
            The boundary point corresponding to the specified boundary condition.
        """
        return self.base.boundarypoint(bc)
    
    def boundarysolution(self, bc):
        """Returns the solution at the boundary for the given boundary condition.
        Parameters
        ---------- 
        bc : str
            The boundary condition to evaluate. Can be "In" for horizon or "Up" for infinity.
        Returns
        -------
        float
            The solution at the boundary corresponding to the specified boundary condition.
        """
        return self.base.boundarysolution(bc)
    
    def boundaryderivative(self, bc):
        """Returns the derivative at the boundary for the given boundary condition.
        Parameters
        ----------
        bc : str
            The boundary condition to evaluate. Can be "In" for horizon or "Up" for infinity.
        Returns
        -------
        float
            The derivative at the boundary corresponding to the specified boundary condition.
        """
        return self.base.boundaryderivative(bc)
    
    def radialsolution(self, bc, pos):
        """Returns the solution at the radial point for the given homogeneous solution and position.
        Parameters
        ----------
        bc : str
            The homogeneous solution to evaluate. Can be "In" for horizon solution or "Up" for infinity solution.
        pos : float
            The position at which to evaluate the solution.
        Returns
        -------
        float
            The solution at the radial point corresponding to the specified boundary condition and position.
        """
        return self.base.solution(bc, pos)
    
    def radialderivative(self, bc, pos):
        """Returns the derivative at the radial point for the given homogeneous solution and position.
        Parameters
        ----------
        bc : str
            The homogeneous solution to evaluate. Can be "In" for horizon solution or "Up" for infinity solution.
        pos : float
            The position at which to evaluate the solution.
        Returns
        -------
        float
            The solution at the radial point corresponding to the specified boundary condition and position.
        """
        return self.base.derivative(bc, pos)
    
    def radialderivative2(self, bc, pos):
        """Returns the second derivative at the radial point for the given homogeneous solution and position.
        Parameters
        ----------
        bc : str
            The homogeneous solution to evaluate. Can be "In" for horizon solution or "Up" for infinity solution.
        pos : float
            The position at which to evaluate the solution.
        Returns
        -------
        float
            The solution at the radial point corresponding to the specified boundary condition and position.
        """
        return self.base.derivative2(bc, pos)

    def radialsolutions(self, bc):
        """Returns a homogeneous solution at all radial points.
        Parameters
        ----------
        bc : str
            The homogeneous solution to evaluate. Can be "In" for horizon solution or "Up" for infinity solution.
        Returns
        -------
        numpy.ndarray
            A numpy array of solutions at all radial points corresponding to the specified boundary condition.
        """
        return np.array([self.base.solution(bc, i) for i in range(self.nsamples)])
    
    def radialderivatives(self, bc):
        """Returns the radial derivative at all radial points.
        Parameters
        ----------
        bc : str
            The homogeneous solution to evaluate. Can be "In" for horizon solution or "Up" for infinity solution.
        Returns
        -------
        numpy.ndarray
            A numpy array of solutions at all radial points corresponding to the specified boundary condition.
        """
        return np.array([self.base.derivative(bc, i) for i in range(self.nsamples)])
    
    def radialderivatives2(self, bc):
        """Returns the second radial derivative at all radial points.
        Parameters
        ----------
        bc : str
            The homogeneous solution to evaluate. Can be "In" for horizon solution or "Up" for infinity solution.
        Returns
        -------
        numpy.ndarray
            A numpy array of solutions at all radial points corresponding to the specified boundary condition.
        """
        return np.array([self.base.derivative2(bc, i) for i in range(self.nsamples)])
    
    def __call__(self, bc, deriv = 0):
        """Returns the solutions, first derivatives, or second derivatives at all radial points for the given boundary condition.

        Parameters
        ----------
        bc : str
            The homogeneous solution to evaluate. Can be "In" for horizon solution or "Up" for infinity solution.
        deriv : int, optional
            Specifies which derivative to return: 0 for solutions, 1 for first derivatives, and 2 for second derivatives.
            Default is 0 (solutions).
        Returns
        -------
        numpy.ndarray
            A numpy array of solutions, first derivatives, or second derivatives at all radial points corresponding to the specified boundary condition.
        Raises
        ------
        ValueError
            If `deriv` is not 0, 1, or 2.
        """ 
        if deriv == 0:
            return self.radialsolutions(bc)
        elif deriv == 1:
            return self.radialderivatives(bc)
        elif deriv == 2:
            return self.radialderivatives2(bc)
        else:
            raise ValueError("RadialTeukolsky only solves up to the second derivative")


from pybhpt.radial import renormalized_angular_momentum, RadialTeukolsky
import pytest
import numpy as np

test_data_renormalized_angular_momentum = [
    ([-2, 5, 2, 0.99, 2.1], -2.45191080799969j),
    ([-2, 15, 2, 0.99, 2.1], 14.855326709178254),
    ([-2, 20, 2, 0.99, 6.5], -5.953247813925858j),
    ([-2, 30, 2, 0.99, 6.5], 29.78319684515244)
]

@pytest.mark.parametrize("params, expected", test_data_renormalized_angular_momentum)
def test_renormalized_angular_momentum(params, expected):
    nu = renormalized_angular_momentum(*params)
    assert np.isclose(np.cos(2.*np.pi*nu), np.cos(2.*np.pi*expected), rtol=1e-9, atol = 0), f"Expected {expected}, got {nu} for parameters {params}"

def test_radial_teukolsky_init():
    with pytest.raises(ValueError):
        RadialTeukolsky(0, 10, 2, 1.1, 0.1, [5.])  # invalid a
    with pytest.raises(ValueError):
        RadialTeukolsky(-1, 1, 2, 0.5, 0.1, [5.])  # invalid j
    with pytest.raises(ValueError):
        RadialTeukolsky(2, 2, 2, 0.5, 0.1, [1.])  # invalid r
    with pytest.raises(AttributeError):
        RadialTeukolsky(0.5, 2, 2, 0.5, 0.1, 10.) # invalid r type

def test_radial_teukolsky_solve():
    with pytest.raises(ValueError):
        Rt = RadialTeukolsky(0, 10, 2, 0.5, 0.1, [5.])
        Rt.solve("HYPERBOLIC", "None")  # invalid method
    with pytest.raises(ValueError):
        Rt = RadialTeukolsky(0.5, 2, 2, 0.5, 0.1, [5.])
        Rt.solve("AUTO", "Down") # invalid boundary condition

@pytest.mark.parametrize("method", ["AUTO", "MST", "ASYMP", "HBL", "GSN", "TEUK"])
def test_radial_teukolsky_solve_auto(method):
    r = np.linspace(2.1, 100, 100)
    Rt = RadialTeukolsky(-2, 2, 2, 0.5, 0.1, r)
    Rt.solve(method, "None")
    assert Rt.radialsolutions('Up').shape == r.shape, f"Radial Up solutions shape mismatch for {method} method"
    assert Rt.radialsolutions('In').shape == r.shape, f"Radial In solutions shape mismatch for {method} method"


from cybhpt_full import KerrGeodesic as KerrGeodesicCython
from cybhpt_full import kerr_geo_V01, kerr_geo_V02, kerr_geo_V11, kerr_geo_V22, kerr_geo_V31, kerr_geo_V32
from cybhpt_full import kerr_mino_frequencies_wrapper, kerr_orbital_constants_wrapper
import numpy as np

def kerrgeo_Vt_radial(a, En, Lz, Q, r):
    return kerr_geo_V01(a, En, Lz, Q, r)

def kerrgeo_Vt_polar(a, En, Lz, Q, theta):
    return kerr_geo_V02(a, En, Lz, Q, theta)

def kerrgeo_Vr(a, En, Lz, Q, r):
    return kerr_geo_V11(a, En, Lz, Q, r)

def kerrgeo_Vtheta(a, En, Lz, Q, theta):
    return kerr_geo_V22(a, En, Lz, Q, theta)

def kerrgeo_Vphi_radial(a, En, Lz, Q, r):
    return kerr_geo_V31(a, En, Lz, Q, r)

def kerrgeo_Vphi_polar(a, En, Lz, Q, theta):
    return kerr_geo_V32(a, En, Lz, Q, theta)

def kerr_mino_frequencies(a, p, e, x):
    """
    Returns the Mino frequencies of a Kerr geodesic.

    Parameters
    ----------
    a : float
        The black hole spin parameter.
    p : float
        The semilatus rectum of the orbit.
    e : float
        The eccentricity of the orbit.
    x : float
        The inclination of the orbit.
    
    Returns
    -------
    numpy.ndarray
        The Mino time frequencies of the orbit.
    """
    return kerr_mino_frequencies_wrapper(a, p, e, x)

def kerr_fundamental_frequencies(a, p, e, x):
    """
    Returns the fundamental (time) frequencies of a Kerr geodesic.
    
    Parameters
    ----------
    a : float
        The black hole spin parameter.
    p : float
        The semilatus rectum of the orbit.
    e : float
        The eccentricity of the orbit.
    x : float
        The inclination of the orbit.
    
    Returns
    -------
    numpy.ndarray
        The fundamental frequencies of the orbit.
    """
    Ups = kerr_mino_frequencies(a, p, e, x)
    return Ups[1:]/Ups[0]

def kerr_orbital_constants(a, p, e, x):
    """
    Returns the orbital constants of a Kerr geodesic (En, Lz, Qc).

    Parameters
    ----------
    a : float
        The black hole spin parameter.
    p : float
        The semilatus rectum of the orbit.
    e : float
        The eccentricity of the orbit.
    x : float
        The inclination of the orbit. 

    Returns
    -------
    numpy.ndarray
        The orbital constants (En, Lz, Qc) of the orbit.
    """
    return kerr_orbital_constants_wrapper(a, p, e, x)

def is_power_of_two(n: int) -> bool:
    return n > 0 and (n & (n - 1)) == 0

class KerrGeodesic:
    """
    Class that produces a Kerr geodesic given the parameters of the orbit.
    This class is a wrapper around the Cython implementation of the Kerr geodesic
    and provides a Python interface to the underlying C++ code.
    Parameters
    ----------
    a : float
        The black hole spin parameter.
    p : float
        The semilatus rectum of the orbit.
    e : float
        The eccentricity of the orbit.
    x : float
        The inclination of the orbit.
    nsamples : int
        The number of samples to use for the geodesic. Must be a power of two. Default is 256.

    Attributes
    ----------
    blackholespin : float
        The black hole spin parameter.
    semilatusrectum : float
        The semilatus rectum of the orbit.
    eccentricity : float
        The eccentricity of the orbit.
    inclination : float
        The inclination of the orbit.
    orbitalenergy : float
        The orbital energy En of the orbit.
    orbitalangularmomentum : float
        Th z-component of the orbital angular momentum Lz of the orbit.
    carterconstant : float
        The Carter constant Qc of the orbit.
    orbitalconstants : numpy.ndarray
        The orbital constants (En, Lz, Qc) of the orbit.
    radialroots : numpy.ndarray
        The roots of the radial equation.
    polarroots : numpy.ndarray
        The roots of the polar equation.
    minofrequencies : numpy.ndarray
        The orbital frequencies with respect to Mino time.
    timefrequencies : numpy.ndarray
        The orbital frequencies with respect to the time coordinate.
    frequencies : numpy.ndarray
        The (coordinate time) frequencies of the orbit.
    carterfrequencies : numpy.ndarray
        The frequencies for computing Carter constant fluxes.
    timeradialfourier : numpy.ndarray
        The Fourier coefficients of coordinate time with respect to the radial Mino phase.
    timepolarfourier : numpy.ndarray
        The Fourier coefficients of coordinate time with respect to the polar Mino phase.
    radialfourier : numpy.ndarray
        The Fourier coefficients of radial position with respect to the radial Mino phase.
    polarfourier : numpy.ndarray
        The Fourier coefficients of polar position with respect to the polar Mino phase.
    azimuthalradialfourier : numpy.ndarray
        The Fourier coefficients of azimuthal position with respect to the radial Mino phase.
    azimuthalpolarfourier : numpy.ndarray
        The Fourier coefficients of azimuthal position with respect to the polar Mino phase.  
    """
    def __init__(self, a, p, e, x, nsamples = 2**8):
        if a < 0 or a > 1:
            raise ValueError(f"Black hole spin parameter {a} must be in the range [0, 1].")
        if not is_power_of_two(nsamples):
            raise ValueError(f"Number of samples {nsamples} must be a power of 2.")

        self.base = KerrGeodesicCython(a, p, e, x, nsamples)
        """The base class that contains the Cython implementation of the Kerr geodesic."""
        self.timeradial = self.base.get_time_accumulation(1)
        self.timepolar = self.base.get_time_accumulation(2)
        self.radialpoints = self.base.get_radial_points()
        self.polarpoints = self.base.get_polar_points()
        self.azimuthalradial = self.base.get_azimuthal_accumulation(1)
        self.azimuthalpolar = self.base.get_azimuthal_accumulation(2)
        self.nsamples = self.base.nsamples

        if np.isnan(self.frequencies).any():
            raise ValueError(f"Orbital parameters (a, p, e, x) = {self.apex} do not represent a valid bound non-plunging orbit.")

    @property
    def blackholespin(self):
        """
        The black hole spin parameter.
        """
        return self.base.blackholespin
    
    @property
    def semilatusrectum(self):
        """
        The semilatus rectum of the orbit.
        """
        return self.base.semilatusrectum
    
    @property
    def eccentricity(self):
        """
        The eccentricity of the orbit.
        """
        return self.base.eccentricity

    @property
    def inclination(self):
        """
        The cosine of the inclination angle of the orbit with respect to the equatorial plane.
        """
        return self.base.inclination
    
    @property
    def a(self):
        """
        The black hole spin parameter.
        """
        return self.blackholespin
    
    @property
    def p(self):
        """
        The semilatus rectum of the orbit.
        """
        return self.semilatusrectum
    
    @property
    def e(self):
        """
        The eccentricity of the orbit.
        """
        return self.eccentricity
    
    @property
    def x(self):
        """
        The inclination of the orbit.
        """
        return self.inclination
    
    @property
    def apex(self):
        """
        The parameters of the orbit (a, p, e, x).
        """
        return np.array([self.blackholespin, self.semilatusrectum, self.eccentricity, self.inclination])
    
    @property
    def orbitalparameters(self):
        """
        The parameters of the orbit (a, p, e, x).
        """
        return self.apex

    @property
    def orbitalenergy(self):
        """
        The orbital energy En of the orbit.
        """
        return self.base.orbitalenergy

    @property
    def orbitalangularmomentum(self):
        """
        The z-component of the orbital angular momentum Lz of the orbit.
        """
        return self.base.orbitalangularmomentum

    @property
    def carterconstant(self):
        """
        The Carter constant Qc of the orbit.
        """
        return self.base.carterconstant

    @property
    def orbitalconstants(self):
        """
        The orbital constants (En, Lz, Qc) of the orbit.
        """
        return np.array([self.orbitalenergy, self.orbitalangularmomentum, self.carterconstant])

    @property
    def radialroots(self):
        """
        The roots of the radial equation.
        """
        return self.base.radialroots

    @property
    def polarroots(self):
        """
        The roots of the polar equation.
        """
        return self.base.polarroots

    @property
    def minofrequencies(self):
        """
        The orbital frequencies with respect to Mino time.
        """
        return self.base.minofrequencies

    @property
    def timefrequencies(self):
        """
        The orbital frequencies with respect to the time coordinate.
        """
        return self.base.timefrequencies

    @property
    def frequencies(self):
        """
        The (coordinate time) frequencies of the orbit.
        """
        return self.timefrequencies

    @property
    def carterfrequencies(self):
        """
        The frequencies for computing Carter constant fluxes.
        """
        return self.base.carterfrequencies
    
    @property
    def timeradialfourier(self):
        """
        The Fourier coefficients of coordinate time with respect to the radial Mino phase.
        """
        return self.base.get_time_coefficients(1)
    
    @property
    def timepolarfourier(self):
        """
        The Fourier coefficients of coordinate time with respect to the polar Mino phase.
        """
        return self.base.get_time_coefficients(2)
    
    @property
    def radialfourier(self):
        """
        The Fourier coefficients of radial position with respect to the radial Mino phase.
        """
        return self.base.get_radial_coefficients()
    
    @property
    def polarfourier(self):
        """
        The Fourier coefficients of polar position with respect to the polar Mino phase.
        """
        return self.base.get_polar_coefficients()
    
    @property
    def azimuthalradialfourier(self):
        """
        The Fourier coefficients of azimuthal position with respect to the radial Mino phase.
        """
        return self.base.get_azimuthal_coefficients(1)
    
    @property
    def azimuthalpolarfourier(self):
        """
        The Fourier coefficients of azimuthal position with respect to the polar Mino phase.
        """
        return self.base.get_azimuthal_coefficients(2)
    
    def mode_frequency(self, m, k, n):
        """
        Returns the frequency of the mode with azimuthal number m, polar number k, and radial number n.
        
        Parameters
        ----------
        m : int
            The azimuthal number of the mode.
        k : int
            The polar number of the mode.
        n : int
            The radial number of the mode.
        
        Returns
        -------
        float
            The frequency of the mode with azimuthal number m, polar number k, and radial number n.
        """
        return np.dot(np.array([n, k, m]), (self.frequencies))
    
    def minotime(self, t):
        """
        Function that returns the Mino time for a given Boyer-Lindquist time.

        Parameters
        ----------
        t : float
            The Boyer-Lindquist time.
        
        Returns
        -------
        float
            The Mino time for the given Boyer-Lindquist time.
        """
        return self.base.mino_time(t)
    
    def __call__(self, la):
        """
        Function that returns the position vector for a given Mino time value.
        Parameters
        ----------
        la : float or numpy.ndarray
            The Mino time value(s). If a numpy array is provided, the function will return a numpy array of the same shape.
        Returns 
        -------
        numpy.ndarray
            The position vector for the given Mino time value(s).
        """
        if isinstance(la, np.ndarray) or isinstance(la, list):
            return self.base.position_vec(np.array(la))
        else:
            return self.base.position(la)


from libcpp.vector cimport vector
from libcpp.complex cimport complex as cpp_complex
from cython.operator cimport dereference
from libcpp.string cimport string

include "flux_wrap.pyx"

cdef extern from "redshift.hpp":
    cdef cppclass RedshiftCoefficientsCPP "RedshiftCoefficients":
        RedshiftCoefficientsCPP(Gauge gauge, GeodesicSource &geo)
        cpp_complex[double] getComponent(int Ni, int ai, int bi, int ci, int di, int jru, int jzu)

cdef extern from "metriccoeffs.hpp":
    cpp_complex[double] metric_coefficient_ORG(int ai, int bi, int nt, int nr, int nz, int nphi, double a, double r, double z)
    cpp_complex[double] metric_coefficient_IRG(int ai, int bi, int nt, int nr, int nz, int nphi, double a, double r, double z)

cdef extern from "metric.hpp":
    cdef cppclass SphericalHarmonicCouplingCPP "SphericalHarmonicCoupling":
        SphericalHarmonicCouplingCPP(int lmax, int m)
        void generateCouplings()
        int getAzimuthalModeNumber()

        double getZCouplingCoefficient(int n, int i, int l)
        double getZCouplingCoefficientNoCheck(int n, int i, int l)
        int getMinZCouplingModeNumber()
        int getMaxZCouplingModeNumber()

        double getDerivativeCouplingCoefficient(int n, int i, int l)
        double getDerivativeCouplingCoefficientNoCheck(int n, int i, int l)
        int getMinDerivativeCouplingModeNumber()
        int getMaxDerivativeCouplingModeNumber()


def metric_coefficients_cython_S4dagger(int ai, int bi, int nt, int nr, int nz, int nphi, double a, double r, double z):
    return metric_coefficient_ORG(ai, bi, nt, nr, nz, nphi, a, r, z)

def metric_coefficients_cython_S0dagger(int ai, int bi, int nt, int nr, int nz, int nphi, double a, double r, double z):
    return metric_coefficient_IRG(ai, bi, nt, nr, nz, nphi, a, r, z)

cdef class RedshiftCoefficients:
    cdef RedshiftCoefficientsCPP *huucpp

    def __cinit__(self, unicode gauge, KerrGeodesic geo):
        self.huucpp = new RedshiftCoefficientsCPP(str_to_gauge(gauge), dereference(geo.geocpp))

    def __dealloc__(self):
        del self.huucpp

    def __call__(self, int Ni, int ai, int bi, int ci, int di, int jr, int jz):
        return self.huucpp.getComponent(Ni, ai, bi, ci, di, jr, jz)

cdef class SphericalHarmonicCoupling:
    cdef SphericalHarmonicCouplingCPP *cpp

    def __cinit__(self, int lmax, int m):
        self.cpp = new SphericalHarmonicCouplingCPP(lmax, m)
        self.cpp.generateCouplings()

    def __dealloc__(self):
        del self.cpp

    @property
    def azimuthalmode(self):
        return self.cpp.getAzimuthalModeNumber()

    def zcouplingcoefficient(self, int n, int i, int l):
        return self.cpp.getZCouplingCoefficient(n, i, l)

    def dzcouplingcoefficient(self, int n, int i, int l):
        return self.cpp.getDerivativeCouplingCoefficient(n, i, l)




from cybhpt_full import RedshiftCoefficients as RedshiftCoefficientsCython
from cybhpt_full import SphericalHarmonicCoupling as SphericalHarmonicCouplingCython

class RedshiftCoefficients:
    def __init__(self, gauge, geo):
        self.base = RedshiftCoefficientsCython(gauge, geo.base)

    def __call__(self, Ni, ai, bi, ci, di, jr = 0, jz = 0):
        return self.base(Ni, ai, bi, ci, di, jr, jz)
    
class SphericalHarmonicCoupling:
    def __init__(self, lmax, m):
        self.base = SphericalHarmonicCouplingCython(lmax, m)
        self.azimuthalmode = self.base.azimuthalmode

    def zcouplingcoefficient(self, n, i, l):
        return self.base.zcouplingcoefficient(n, i, l)

    def dzcouplingcoefficient(self, n, i, l):
        return self.base.dzcouplingcoefficient(n, i, l)


from cybhpt_full import HertzMode as HertzModeCython
from cybhpt_full import test_hertz_mode_cython
from cybhpt_full import teuk_to_hertz_ORG, teuk_to_hertz_IRG, teuk_to_hertz_SRG, teuk_to_hertz_ARG
from pybhpt.radial import RadialTeukolsky
import numpy as np

available_gauges = [
    "IRG",
    "ORG",
    "SRG0", 
    "SRG4", 
    "ARG0", 
    "ARG4",
]

def hertz_IRG(Zin, Zup, j, m, k, a, omega, lambdaCH):
    """
    Convert Teukolsky amplitudes to Hertz potential in the IRG gauge.
    """
    return teuk_to_hertz_IRG(Zin, Zup, j, m, k, a, omega, lambdaCH)

def hertz_ORG(Zin, Zup, j, m, k, a, omega, lambdaCH):
    """Convert Teukolsky amplitudes to Hertz potential in the ORG gauge.
    """
    return teuk_to_hertz_ORG(Zin, Zup, j, m, k, a, omega, lambdaCH)

def hertz_SRG(Zin, Zup, j, m, k, a, omega, lambdaCH):
    """Convert Teukolsky amplitudes to Hertz potential in the SRG gauge.
    """
    return teuk_to_hertz_SRG(Zin, Zup, j, m, k, a, omega, lambdaCH)

def hertz_ARG(Zin, Zup, j, m, k, a, omega, lambdaCH):
    """Convert Teukolsky amplitudes to Hertz potential in the ARG gauge.
    """
    return teuk_to_hertz_ARG(Zin, Zup, j, m, k, a, omega, lambdaCH)

def teuk_to_hertz_amplitude(gauge, Zin, Zup, j, m, k, a, omega, lambdaCH):
    if gauge == "IRG":
        return hertz_IRG(Zin, Zup, j, m, k, a, omega, lambdaCH)
    elif gauge == "ORG":
        return hertz_ORG(Zin, Zup, j, m, k, a, omega, lambdaCH)
    elif gauge == "SRG0" or gauge == "SRG4":
        return hertz_SRG(Zin, Zup, j, m, k, a, omega, lambdaCH)
    elif gauge == "ARG0" or gauge == "ARG4":
        return hertz_ARG(Zin, Zup, j, m, k, a, omega, lambdaCH)
    else:
        return (0.j, 0.j)

def test_hertz_mode(j, m, k, n, geo):
    test_hertz_mode_cython(j, m, k, n, geo.base)

def gauge_check(gauge):
    """ Check if the provided gauge is supported. 
    Raises a TypeError if the gauge is not supported.
    """
    if gauge not in available_gauges:
        TypeError("{} is not a supported gauge.".format(gauge))


class HertzMode:
    """
    Class that produces a Hertz potential mode given a Teukolsky object and a gauge.
    This class is a wrapper around the Cython implementation of the Hertz potential
    and provides a Python interface to the underlying C++ code.

    Parameters
    ----------
    teuk : Teukolsky
        The Teukolsky object to be used for the Hertz potential.
    gauge : str
        The gauge to be used for the Hertz potential. Must be one of the following:
        - "IRG"
        - "ORG"
        - "SRG0"
        - "SRG4"
        - "ARG0"
        - "ARG4"

    Attributes
    ----------
    base : HertzModeCython
        The underlying Cython implementation of the Hertz potential mode.
    gauge : str
        The gauge used for the Hertz potential.
    sampleR : int
        The number of radial samples used in the Hertz potential mode solutions
    sampleTh : int
        The number of polar samples used in the Hertz potential mode solutions
    spinweight : int
        The spin weight of the Hertz potential mode.
    spheroidalmode : int
        The spheroidal mode number of the Hertz potential mode.
    azimuthalmode : int
        The azimuthal mode number of the Hertz potential mode.
    radialmode : int
        The radial mode number of the Hertz potential mode.
    polarmode : int
        The polar mode number of the Hertz potential mode.
    blackholespin : float
        The spin of the black hole associated with the background spacetime.
    frequency : float
        The frequency of the Hertz potential mode.
    horizonfrequency : float
        The frequency of the Hertz potential mode at the horizon.
    eigenvalue : complex
        The spheroidal eigenvalue associated with the Hertz potential mode.
    mincouplingmode : int
        The minimum l-mode used for coupling the spherical and spheroidal harmonics
    maxcouplingmode : int
        The maximum l-mode used for coupling the spherical and spheroidal harmonics
    minscalarcouplingmode : int
        The minimum l-mode used for coupling the scalar harmonics
    maxscalarcouplingmode : int
        The maximum l-mode used for coupling the scalar harmonics
    j : int
        Alias for spheroidalmode.
    m : int
        Alias for azimuthalmode.
    k : int
        Alias for polarmode.
    n : int
        Alias for radialmode.
    omega : float
        Alias for frequency.
    a : float
        Alias for blackholespin.
    
    Properties
    ----------
    couplingcoefficients : np.ndarray
        The coupling coefficients for the Hertz potential mode.
    scalarcouplingcoefficients : np.ndarray
        The scalar coupling coefficients for the Hertz potential mode.
    polarpoints : np.ndarray
        The polar points used in the Hertz potential mode solutions.
    polarsolutions : np.ndarray
        The polar mode solutions of the Hertz potential.
    polarderivatives : np.ndarray
        Derivatives of the polar mode solutions of the Hertz potential.
    polarderivatives2 : np.ndarray
        Second derivatives of the polar mode solutions of the Hertz potential.
    radialpoints : np.ndarray
        The radial points used in the Hertz potential mode solutions.
    radialsolutions : np.ndarray
        The radial mode solutions of the Hertz potential.
    radialderivatives : np.ndarray
        Derivatives of the radial mode solutions of the Hertz potential.
    radialderivatives2 : np.ndarray
        Second derivatives of the radial mode solutions of the Hertz potential.

    Methods
    -------
    solve()
        Solve the Hertz potential mode equations.
    couplingcoefficient(l)
        Returns the coupling coefficient for the given l-mode.
    scalarcouplingcoefficient(l)
        Returns the scalar coupling coefficient for the given l-mode.
    radialpoint(pos)
        Returns the radial point corresponding to the given position.
    radialsolution(bc, pos)
        Returns the radial solution for the given boundary condition and position.
    radialderivative(bc, pos)
        Returns the radial derivative for the given boundary condition and position.
    radialderivative2(bc, pos)
        Returns the second radial derivative for the given boundary condition and position.
    homogeneousradialsolution(bc, pos)
        Returns the homogeneous radial solution for the given boundary condition and position.
    homogeneousradialderivative(bc, pos)
        Returns the homogeneous radial derivative for the given boundary condition and position.
    homogeneousradialderivative2(bc, pos)
        Returns the second homogeneous radial derivative for the given boundary condition and position.
    polarpoint(pos)
        Returns the polar point corresponding to the given position.
    polarsolution(pos)
        Returns the polar solution for the given position.
    polarderivative(pos)
        Returns the polar derivative for the given position.
    polarderivative2(pos)
        Returns the second polar derivative for the given position.
    amplitude(bc)
        Returns the Hertz amplitude for the given boundary condition.
    __call__(r, deriv=0)
        Returns the radial Hertz potential mode evaluated at the given radial values `r`.
        Alternatively, it can return the radial derivative if `deriv` is set to 1 or 2.
        The radial values `r` must lie outside the source region defined by the radial points.
        If r contains values inside the source region, a ValueError is raised.
    """
    def __init__(self, teuk, gauge):
        self.base = HertzModeCython(teuk.base, gauge)
        self.gauge = gauge
        self.sampleR = self.base.radialsamplenumber
        self.sampleTh = self.base.polarsamplenumber

    @property
    def spinweight(self):
        return self.base.spinweight

    @property
    def spheroidalmode(self):
        return self.base.spheroidalmode
    
    @property
    def azimuthalmode(self):
        return self.base.azimuthalmode

    @property
    def radialmode(self):
        return self.base.radialmode

    @property
    def polarmode(self):
        return self.base.spipolarmodenweight

    @property
    def blackholespin(self):
        return self.base.blackholespin
    
    @property
    def frequency(self):
        return self.base.frequency

    @property
    def horizonfrequency(self):
        return self.base.horizonfrequency

    @property
    def eigenvalue(self):
        return self.base.eigenvalue

    @property
    def mincouplingmode(self):
        return self.base.mincouplingmode
    
    @property
    def maxcouplingmode(self):
        return self.base.maxcouplingmode
    
    @property
    def minscalarcouplingmode(self):
        return self.base.minscalarcouplingmode
    
    @property
    def maxscalarcouplingmode(self):
        return self.base.maxscalarcouplingmode

    # some useful aliases
    @property
    def j(self):
        return self.spheroidalmode
    @property
    def m(self):
        return self.azimuthalmode
    @property
    def k(self):
        return self.polarmode
    @property
    def n(self):
        return self.radialmode
    @property
    def omega(self):
        return self.frequency
    @property
    def a(self):
        return self.blackholespin
    
    @property
    def couplingcoefficients(self):
        return self.base.couplingcoefficients
    
    @property
    def scalarcouplingcoefficients(self):
        return self.base.scalarcouplingcoefficients
    
    @property
    def polarpoints(self):
        return self.base.polarpoints
    
    @property
    def polarsolutions(self):
        return self.base.polarsolutions
    
    @property
    def polarderivatives(self):
        return self.base.polarderivatives
    
    @property
    def polarderivatives2(self):
        return self.base.polarderivatives2
    
    @property
    def radialpoints(self):
        return self.base.radialpoints
    
    @property
    def radialsolutions(self):
        return self.base.radialsolutions
    
    @property
    def radialderivatives(self):
        return self.base.radialderivatives
    
    @property
    def radialderivatives2(self):
        return self.base.radialderivatives2
    
    @property
    def amplitudes(self):
        return {"In": self.amplitude('In'), "Up": self.amplitude('Up')}
    
    def solve(self):
        self.base.solve()

    def couplingcoefficient(self, l):
        return self.base.couplingcoefficient(l)
    
    def scalarcouplingcoefficient(self, l):
        return self.base.scalarcouplingcoefficient(l)

    def radialpoint(self, pos):
        return self.base.radialpoint(pos)
    
    def radialsolution(self, bc, pos):
        return self.base.radialsolution(bc, pos)
    
    def radialderivative(self, bc, pos):
        return self.base.radialderivative(bc, pos)
    
    def radialderivative2(self, bc, pos):
        return self.base.radialderivative2(bc, pos)
    
    def homogeneousradialsolution(self, bc, pos):
        return self.base.homogeneousradialsolution(bc, pos)
    
    def homogeneousradialderivative(self, bc, pos):
        return self.base.homogeneousradialderivative(bc, pos)
    
    def homogeneousradialderivative2(self, bc, pos):
        return self.base.homogeneousradialderivative2(bc, pos)
    
    def polarpoint(self, pos):
        return self.base.polarpoint(pos)
    
    def polarsolution(self, pos):
        return self.base.polarsolution(pos)
    
    def polarderivative(self, pos):
        return self.base.polarderivative(pos)
    
    def polarderivative2(self, pos):
        return self.base.polarderivative2(pos)
    
    def amplitude(self, bc):
        return self.base.hertz_amplitude(bc)
    
    def __call__(self, r, deriv = 0):
        """
        Evaluate the radial Hertz potential mode at the given radial values `r`.
        If `deriv` is set to 0, it returns the Hertz potential mode.
        If `deriv` is set to 1, it returns the first radial derivative.
        If `deriv` is set to 2, it returns the second radial derivative.
        
        Parameters
        ----------
        r : array-like
            The radial values at which to evaluate the Hertz potential mode.
        deriv : int, optional
            The order of the radial derivative to compute. Default is 0 (no derivative). 
        Returns
        -------
        numpy.ndarray
            The evaluated Hertz potential mode or its radial derivative at the given radial values `r`.
        Raises
        ------
        ValueError
            If any of the radial values `r` lie within the source region defined by the radialpoints.
        """
        rmin = self.radialpoints[0]
        rmax = self.radialpoints[-1]
        rinner = r[r < rmin]
        router = r[r > rmax]
        if np.any((r >= rmin) & (r <= rmax)):
            raise ValueError("Radial points must lie outside the source region")

        if rinner.shape[0] > 0:
            Rt = RadialTeukolsky(self.spinweight, self.spheroidalmode, self.azimuthalmode, self.blackholespin, self.frequency, rinner)
            Rt.solve(bc='In')
            Rin = Rt('In', deriv=deriv)
        else:
            Rin = np.array([])

        if router.shape[0] > 0:
            Rt = RadialTeukolsky(self.spinweight, self.spheroidalmode, self.azimuthalmode, self.blackholespin, self.frequency, router)
            Rt.solve(bc='Up')
            Rup = Rt('Up', deriv=deriv)
        else:
            Rup = np.array([])

        PsiIn = self.amplitude('In')
        PsiUp = self.amplitude('Up')

        return np.concatenate((PsiIn*Rin, PsiUp*Rup))


from cybhpt_full import metric_coefficients_cython_S4dagger, metric_coefficients_cython_S0dagger
import numpy as np
from pybhpt.hertz import available_gauges
from pybhpt.swsh import Yslm, spin_operator_normalization

def gauge_check(gauge):
    if gauge not in available_gauges:
        TypeError("{} is not a supported gauge.".format(gauge))

S0_gauges = ["IRG", "ARG0", "SRG0"]
S4_gauges = ["ORG", "ARG4", "SRG4"]

def metric_coefficients_S4dagger_ab(ai, bi, nt, nr, nz, nph, a, r, z):
    """Compute the metric coefficients for the reconstructed perturbation associated with the
    S4dagger reconstruction operator.
    
    Parameters
    ----------
    ai, bi : int
        The indices of the tetrad-projected components of the metric perturbation.
    nt, nr, nz, nph : int
        The indices associated with the time, radial, spin, and azimuthal derivative operators.
    a : float
        The black hole spin parameter.
    r : float
        The radial coordinate.
    z : float
        The polar coordinate (cosine of the polar angle).
    
    Returns
    -------
    float
        The computed metric coefficient."""
    return metric_coefficients_cython_S4dagger(ai, bi, nt, nr, nz, nph, a, r, z)

def metric_coefficients_S0dagger_ab(ai, bi, nt, nr, nz, nph, a, r, z):
    """Compute the metric coefficients for the reconstructed perturbation associated with the
    S0dagger reconstruction operator.
    
    Parameters
    ----------
    ai, bi : int
        The indices of the tetrad-projected components of the metric perturbation.
    nt, nr, nz, nph : int
        The indices associated with the time, radial, spin, and azimuthal derivative operators.
    a : float
        The black hole spin parameter.
    r : float
        The radial coordinate.
    z : float
        The polar coordinate (cosine of the polar angle).
    
    Returns
    -------
    float
        The computed metric coefficient."""
    return metric_coefficients_cython_S0dagger(ai, bi, nt, nr, nz, nph, a, r, z)

def metric_coefficients_S0dagger(a, b, c, d, q, rvals, zvals):
    """Compute the metric coefficients for the reconstructed perturbation associated with the
    S0dagger reconstruction operator.
    
    Parameters
    ----------
    a, b, c, d : int
        The indices associated with the time, radial, spin, and azimuthal derivative operators.
    q : float
        The black hole spin parameter.
    rvals : numpy.ndarray
        The radial coordinate.
    zvals : np.ndarray
        The polar coordinate (cosine of the polar angle).
    
    Returns
    -------
    numpy.ndarray
        A 3D array containing the computed metric coefficients for the S0dagger operator."""
    h22 = np.array([[metric_coefficients_S0dagger_ab(2, 2, a, b, c, d, q, r, z) for z in zvals] for r in rvals])
    h24 = np.array([[metric_coefficients_S0dagger_ab(2, 4, a, b, c, d, q, r, z) for z in zvals] for r in rvals])
    h44 = np.array([[metric_coefficients_S0dagger_ab(4, 4, a, b, c, d, q, r, z) for z in zvals] for r in rvals])
    return np.array([2.*h22, h24, h44])

def metric_coefficients_S4dagger(a, b, c, d, q, rvals, zvals):
    """Compute the metric coefficients for the reconstructed perturbation associated with the
    S4dagger reconstruction operator.
    
    Parameters
    ----------
    a, b, c, d : int
        The indices associated with the time, radial, spin, and azimuthal derivative operators.
    q : float
        The black hole spin parameter.
    rvals : numpy.ndarray
        The radial coordinate.
    zvals : np.ndarray
        The polar coordinate (cosine of the polar angle).
    
    Returns
    -------
    numpy.ndarray
        A 3D array containing the computed metric coefficients for the S0dagger operator."""
    h11 = np.array([[metric_coefficients_S4dagger_ab(1, 1, a, b, c, d, q, r, z) for z in zvals] for r in rvals])
    h13 = np.array([[metric_coefficients_S4dagger_ab(1, 3, a, b, c, d, q, r, z) for z in zvals] for r in rvals])
    h33 = np.array([[metric_coefficients_S4dagger_ab(3, 3, a, b, c, d, q, r, z) for z in zvals] for r in rvals])
    return np.array([2.*h11, h13, h33])

def metric_coefficients(gauge, a, b, c, d, q, rvals, zvals):
    """Compute the metric coefficients for the reconstructed perturbation associated with the
    specified gauge.

    Parameters
    ----------
    gauge : str
        The gauge to use for the reconstruction. Must be one of the available gauges.
    a, b, c, d : int
        The indices associated with the time, radial, spin, and azimuthal derivative operators.
    q : float
        The black hole spin parameter.
    rvals : numpy.ndarray
        The radial coordinate.
    zvals : np.ndarray
        The polar coordinate (cosine of the polar angle).

    Returns
    -------
    numpy.ndarray
        A 3D array containing the computed metric coefficients for the specified gauge.
    """
    gauge_check(gauge)
    if gauge in S0_gauges:
        return metric_coefficients_S0dagger(a, b, c, d, q, rvals, zvals)
    else:
        return metric_coefficients_S4dagger(a, b, c, d, q, rvals, zvals)

class MetricCoefficients:
    """
    A class for computing the metric coefficients of the reconstructed perturbation
    associated with the S0dagger or S4dagger reconstruction operator based on the specified gauge.

    Parameters
    ----------
    gauge : str
        The gauge to use for the reconstruction. Must be one of the available gauges.
    q : float
        The black hole spin parameter.
    r : numpy.ndarray
        The radial coordinate values.
    th : numpy.ndarray
        The polar coordinate values (cosine of the polar angle).

    Attributes
    ----------
    gauge : str
        The gauge used for the reconstruction.
    blackholespin : float
        The black hole spin parameter.
    radialpoints : numpy.ndarray
        The radial coordinate values.
    polarpoints : numpy.ndarray
        The polar coordinate values (cosine of the polar angle).
    storedcomponents : dict
        A dictionary mapping pairs of indices (a, b) to the index of the stored component
        in the coefficients array.
    conjugatecomponents : dict
        A dictionary mapping pairs of indices (a, b) to the index of the conjugate component
        in the coefficients array.
    coeffs : numpy.ndarray
        A 7D array containing the computed metric coefficients for the specified gauge.
    zeros : numpy.ndarray
        A 2D array of zeros with the same shape as the radial and polar coordinate arrays.

    Methods
    -------
    hab(a, b, nt, nr, ns, nphi):
        Returns the metric coefficient for the specified indices (a, b) and derivative orders
        (nt, nr, ns, nphi). If the indices are not found in the stored or conjugate components,
        it returns a zero array.
    __call__(a, b, nt, nr, ns, nphi):
        Calls the `hab` method to retrieve the metric coefficient for the specified indices
        and derivative orders.
    """
    def __init__(self, gauge, q, r, th):
        gauge_check(gauge)
        self.gauge = gauge
        rsamples = r.shape[0]
        zsamples = th.shape[0]
        self.blackholespin = q
        self.radialpoints = r
        self.polarpoints = th
        if gauge in S0_gauges:
            self.storedcomponents = {(2, 2): 0, (2, 4): 1, (4, 4): 2}
            self.conjugatecomponents = {(2, 3): 1, (3, 3): 2}
        else:
            self.storedcomponents = {(1, 1): 0, (1, 3): 1, (3, 3): 2}
            self.conjugatecomponents = {(1, 4): 1, (4, 4): 2}
        z = np.cos(th)
        z[np.abs(z) < 1.e-15] = 0.
        self.coeffs = np.zeros((3, 3, 3, 3, 3, rsamples, zsamples), dtype=np.complex128)
        self.zeros = np.zeros((rsamples, zsamples), dtype=np.complex128)
        for ai in range(3):
            for bi in range(3):
                for ci in range(3):
                    for di in range(3):
                        self.coeffs[ai, bi, ci, di] = metric_coefficients(self.gauge, ai, bi, ci, di, q, r, z)

    def hab(self, a, b, nt, nr, ns, nphi):
        """
        Returns the metric coefficient for the specified indices (a, b) and derivative orders
        (nt, nr, ns, nphi). If the indices are not found in the stored or conjugate components,
        it returns a zero array.

        Parameters
        ----------
        a, b : int
            The indices of the tetrad-projected components of the metric perturbation.
        nt, nr, ns, nphi : int
            The indices associated with the time, radial, spin, and azimuthal derivative operators.

        Returns
        -------
        numpy.ndarray
            The computed metric coefficient for the specified indices and derivative orders.
        """
        if b < a:
            atemp = a
            a = b
            b = atemp
        if (a, b) in self.storedcomponents.keys():
            return self.coeffs[nt, nr, ns, nphi][self.storedcomponents[(a, b)]]
        elif (a, b) in self.conjugatecomponents.keys():
            return np.conj(self.coeffs[nt, nr, ns, nphi][self.conjugatecomponents[(a, b)]])
        else:
            return self.zeros

    def __call__(self, a, b, nt, nr, ns, nphi):
        """ Calls the `hab` method to retrieve the metric coefficient for the specified indices
        and derivative orders.

        Parameters
        ----------
        a, b : int
            The indices of the tetrad-projected components of the metric perturbation.
        nt, nr, ns, nphi : int
            The indices associated with the time, radial, spin, and azimuthal derivative operators.

        Returns
        -------
        numpy.ndarray
            The computed metric coefficient for the specified indices and derivative orders.
        """
        return self.hab(a, b, nt, nr, ns, nphi)
    
def tetrad_project_l(a, r, z, mu):
    if mu == 0:
        return -1
    elif mu == 1:
        sigma = r**2 + a**2*z**2
        delta = r**2 - 2.*r + a**2
        return sigma/delta
    elif mu == 2:
        return 0.
    elif mu == 3:
        return a*(1. - z**2)
    else:
        return 0.
    
def tetrad_project_n(a, r, z, mu):
    sigma = r**2 + a**2*z**2
    delta = r**2 - 2.*r + a**2
    if mu == 0:
        return -0.5*delta/sigma
    elif mu == 1:
        return -0.5
    elif mu == 2:
        return 0.
    elif mu == 3:
        return 0.5*delta/sigma*a*(1. - z**2)
    else:
        return 0.
    
def tetrad_project_m(a, r, z, mu):
    rhobar = -1./(r + 1j*a*z)
    sigma = r**2 + a**2*z**2
    pref = - rhobar*np.sqrt(0.5*(1. - z**2))
    if mu == 0:
        return -1j*pref*a
    elif mu == 1:
        return 0.
    elif mu == 2:
        return -pref*sigma/(1. - z**2)
    elif mu == 3:
        return 1j*pref*(r**2 + a**2)
    else:
        return 0.
    
def kinnersley_tetrad_covector(b, q, r, z, mu):
    if b == 1:
        return tetrad_project_l(q, r, z, mu)
    elif b == 2:
        return tetrad_project_n(q, r, z, mu)
    elif b == 3:
        return tetrad_project_m(q, r, z, mu)
    elif b == 4:
        return np.conj(tetrad_project_m(q, r, z, mu))
    else:
        return 0.
    
def hmunu_BL(gauge, mu, nu, q, r, hab):
    gauge_check(gauge)
    if gauge in S0_gauges:
        e1mu = kinnersley_tetrad_covector(1, q, r, 0, mu)
        e1nu = kinnersley_tetrad_covector(1, q, r, 0, nu)
        e3mu = kinnersley_tetrad_covector(3, q, r, 0, mu)
        e3nu = kinnersley_tetrad_covector(3, q, r, 0, nu)
        h22 = e1mu*e1nu*hab[0]
        h24 = -e1mu*e3nu*hab[1] - e1nu*e3mu*hab[1]
        h44 = e3mu*e3nu*hab[2]
        return (h22.real + 2.*h24.real + 2.*h44.real)
    else:
        e1mu = kinnersley_tetrad_covector(2, q, r, 0, mu)
        e1nu = kinnersley_tetrad_covector(2, q, r, 0, nu)
        e3mu = kinnersley_tetrad_covector(4, q, r, 0, mu)
        e3nu = kinnersley_tetrad_covector(4, q, r, 0, nu)
        h11 = e1mu*e1nu*hab[0]
        h13 = -e1mu*e3nu*hab[1] - e1nu*e3mu*hab[1]
        h33 = e3mu*e3nu*hab[2]
        return (h11.real + 2.*h13.real + 2.*h33.real)
